MODULE icb_oce

   !!======================================================================
   !!                       ***  MODULE  icb_oce  ***
   !! Ocean physics:  declare variables for iceberg tracking
   !!======================================================================
   !! History : 3.3.1 !  2010-01  (Martin&Adcroft) Original code
   !!            -    !  2011-03  (Madec)          Part conversion to NEMO form
   !!            -    !                            Removal of mapping from another grid
   !!            -    !  2011-04  (Alderson)       Extensive rewrite
   !!            -    !                            Split into separate modules
   !!----------------------------------------------------------------------
   !!
   !! Track Icebergs as Lagrangian objects within the model domain
   !! Interaction with the other model variables through 'icebergs_gridded'
   !!
   !! A single iceberg is held as an instance of type 'iceberg'
   !! This type defines a linked list, so each instance contains a pointer to the previous and next
   !! icebergs in the list
   !!
   !! Type 'icebergs' is a convenience container for all relevant arrays
   !! It contains one pointer to an 'iceberg' instance representing all icebergs in the processor
   !!
   !! Each iceberg has a position represented as a real cartesian coordinate which is fractional
   !! grid cell, centred on T points; so an iceberg position of (1.0,1.0) lies exactly on the first
   !! T point and the T cell spans 0.5 to 1.5 in each direction
   !!
   !! Each iceberg is assigned a unique id even in MPI
   !! This consists of an array of integers: the first element is used to label, the second and
   !! subsequent elements are used to count the number of times the first element wraps around 
   !! all possible values within the valid size for this datatype.
   !! Labelling is done by starting the first label in each processor (even when only one)
   !! as narea, and then incrementing by jpnij (i.e. the total number of processors.
   !! This means that the source processor for each iceberg can be identified by arithmetic
   !! modulo jpnij.
   !! 
   !!----------------------------------------------------------------------

   USE par_oce         ! ocean parameters
   USE fldread

   IMPLICIT NONE
   PUBLIC

   ! Global constants
   INTEGER, PUBLIC, PARAMETER :: nclasses = 10 ! Number of ice bergs classes
   INTEGER, PUBLIC, PARAMETER :: nkounts  =  3 ! Number of integers combined for unique naming

   !! icebergs_gridded type allows for interaction between the set of icebergs and the model grid
   TYPE, PUBLIC :: icebergs_gridded
      REAL(wp), DIMENSION(:,:)  , POINTER :: calving => NULL()     ! Calving mass rate [frozen runoff] (kg/s) (into stored ice)
      REAL(wp), DIMENSION(:,:)  , POINTER :: calving_hflx=>NULL()  ! Calving heat flux [heat content of calving] (W/m2)
      REAL(wp), DIMENSION(:,:)  , POINTER :: floating_melt=>NULL() ! Net melting rate to icebergs + bits (kg/s/m^2)
      INTEGER , DIMENSION(:,:)  , POINTER :: maxclass=>NULL()      ! maximum class number at calving source point
      REAL(wp), DIMENSION(:,:)  , POINTER :: tmp=>NULL()           ! Temporary work space
      REAL(wp), DIMENSION(:,:,:), POINTER :: stored_ice=>NULL()    ! Accumulated ice mass flux at calving locations (kg)
      REAL(wp), DIMENSION(:,:)  , POINTER :: stored_heat=>NULL()   ! Heat content of stored ice (J)
   END TYPE icebergs_gridded

   !! point type is just a convenience storage type for properties of an individual iceberg
   TYPE, PUBLIC ::   point
      INTEGER  :: year
      REAL(wp) :: xi, yj              ! iceberg coordinates in the (i,j) referential (global)
      REAL(wp) :: e1, e2              ! horizontal scale factors at the iceberg position
      REAL(wp) :: lon, lat, day
      REAL(wp) :: mass, thickness, width, length, uvel, vvel
      REAL(wp) :: uo, vo, ui, vi, ua, va, ssh_x, ssh_y, sst, cn, hi
      REAL(wp) :: mass_of_bits, heat_density
   END TYPE point

   !! iceberg type is really just a linked list
   TYPE, PUBLIC :: iceberg
      ! pointers to previous and next unique icebergs in linked list
      TYPE(iceberg), POINTER      :: prev=>NULL(), next=>NULL()
      ! variables which do not change for this iceberg
      INTEGER, DIMENSION(nkounts) :: number
      REAL(wp)                    :: mass_scaling
      ! variables which change with time are held in a separate type
      TYPE(point), POINTER        :: current_point => NULL()
   END TYPE iceberg

   !! master instances of gridded and linked list iceberg types
   TYPE(icebergs_gridded), POINTER ::   berg_grid
   TYPE(iceberg)         , POINTER ::   first_berg   => NULL()

   !! parameters controlling iceberg characteristics and modelling
   REAL(wp)                        ::   berg_dt                    ! Time-step between iceberg CALLs (should make adaptive?)
   INTEGER                         ::   current_year
   REAL(wp)                        ::   current_yearday            ! 1.00-365.99
   REAL(wp), DIMENSION(:), POINTER ::   initial_width, initial_length
   LOGICAL                         ::   restarted_bergs=.false.          ! Indicate whether we read state from a restart or not
   !                                                           ! arbitrary numbers for diawri entry
   REAL(wp), DIMENSION(nclasses), PUBLIC ::   class_num=(/ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 /)

   ! Extra arrays with bigger halo, needed when interpolating forcing onto iceberg position
   ! particularly for MPP when iceberg can lie inside T grid but outside U, V, or f grid
   REAL(wp), PUBLIC, DIMENSION(:,:), ALLOCATABLE :: uo_e
   REAL(wp), PUBLIC, DIMENSION(:,:), ALLOCATABLE :: vo_e
   REAL(wp), PUBLIC, DIMENSION(:,:), ALLOCATABLE :: ff_e
   REAL(wp), PUBLIC, DIMENSION(:,:), ALLOCATABLE :: ua_e
   REAL(wp), PUBLIC, DIMENSION(:,:), ALLOCATABLE :: va_e
#if defined key_lim2 || defined key_lim3
   REAL(wp), PUBLIC, DIMENSION(:,:), ALLOCATABLE :: ui_e
   REAL(wp), PUBLIC, DIMENSION(:,:), ALLOCATABLE :: vi_e
   REAL(wp), PUBLIC, DIMENSION(:,:), ALLOCATABLE :: ssh_e
#endif

   !!gm almost all those PARAM ARE defined in NEMO
   REAL(wp), PUBLIC, PARAMETER :: rho_ice      = 916.7_wp   ! Density of fresh ice @ 0oC (kg/m^3)
   REAL(wp), PUBLIC, PARAMETER :: rho_water    = 999.8_wp   ! Density of fresh water @ 0oC (kg/m^3)
   REAL(wp), PUBLIC, PARAMETER :: rho_air      = 1.1_wp     ! Density of air @ 0oC (kg/m^3) ???
   REAL(wp), PUBLIC, PARAMETER :: rho_seawater = 1025._wp   ! Approx. density of surface sea water @ 0oC (kg/m^3)
   !!gm end
   REAL(wp), PUBLIC, PARAMETER :: Cd_av = 1.3_wp      ! (Vertical) Drag coefficient between bergs and atmos (?)
   REAL(wp), PUBLIC, PARAMETER :: Cd_ah = 0.0055_wp   ! (Horizontal) Drag coefficient between bergs and atmos (?)
   REAL(wp), PUBLIC, PARAMETER :: Cd_wv = 0.9_wp      ! (Vertical) Drag coefficient between bergs and ocean (?)
   REAL(wp), PUBLIC, PARAMETER :: Cd_wh = 0.0012_wp   ! (Horizontal) Drag coefficient between bergs and ocean (?)
   REAL(wp), PUBLIC, PARAMETER :: Cd_iv = 0.9_wp      ! (Vertical) Drag coefficient between bergs and sea-ice (?)
   !TOM> no horizontal drag for sea ice! real, PARAMETER :: Cd_ih=0.0012 ! (Horizontal) Drag coefficient between bergs and sea-ice (?)

   !                                           !** namberg namelist parameters (and defaults) **
   LOGICAL,  PUBLIC  ::   ln_bergdia        = .true.    ! Calculate budgets
   INTEGER,  PUBLIC  ::   nn_verbose_level       = 0         ! Turn on debugging when level > 0
   INTEGER,  PUBLIC  ::   nn_test_icebergs = 0    ! Create icebergs in absence of a restart file from the supplied class number
   REAL(wp), PUBLIC, DIMENSION(4) ::   rn_test_box = (/ 0._wp, 1._wp, 0._wp, 1._wp /)
                                                         ! lon1,lon2,lat1,lat2 box to create them in
   INTEGER,   PUBLIC ::   nn_sample_rate           = 0       ! Timesteps between sampling of position for trajectory storage
   INTEGER,   PUBLIC ::   nn_verbose_write              = 15      ! timesteps between verbose messages
   REAL(wp),  PUBLIC ::   rn_rho_bergs                  = 850._wp ! Density of icebergs
   REAL(wp),  PUBLIC ::   rn_LoW_ratio                  = 1.5_wp  ! Initial ratio L/W for newly calved icebergs
   REAL(wp),  PUBLIC ::   rn_bits_erosion_fraction = 0.      ! Fraction of erosion melt flux to divert to bergy bits
   REAL(wp),  PUBLIC ::   rn_sicn_shift                 = 0._wp   ! Shift of sea-ice concentration in erosion flux modulation (0<sicn_shift<1)
   LOGICAL,   PUBLIC ::   ln_operator_splitting     = .true.  ! Use first order operator splitting for thermodynamics
   LOGICAL,   PUBLIC ::   ln_passive_mode               = .false. ! iceberg - ocean decoupling
   LOGICAL,   PUBLIC ::   ln_time_average_weight        = .false. ! Time average the weight on the ocean        !!gm I don't understand that !
   REAL(wp),  PUBLIC ::   rn_speed_limit                = 0._wp   ! CFL speed limit for a berg
   !
   !                                                           ! Mass thresholds between iceberg classes (kg)
   REAL(wp), DIMENSION(nclasses), PUBLIC ::   rn_initial_mass=(/8.8e7, 4.1e8, 3.3e9, 1.8e10, 3.8e10, 7.5e10, 1.2e11, &
                                                             2.2e11, 3.9e11, 7.4e11/)
   !                                              ! Fraction of calving to apply to this class (non-dim)
   REAL(wp), DIMENSION(nclasses), PUBLIC ::   rn_distribution=(/0.25, 0.12, 0.15, 0.18, 0.12, 0.07, 0.03, 0.03, 0.03, 0.02/)
   !                                              ! Ratio between effective and real iceberg mass (non-dim)
   REAL(wp), DIMENSION(nclasses), PUBLIC ::   rn_mass_scaling=(/2000, 200, 50, 20, 10, 5, 2, 1, 1, 1/)
   !                                              ! Total thickness of newly calved bergs (m)
   REAL(wp), DIMENSION(nclasses), PUBLIC ::   rn_initial_thickness=(/40., 67., 133., 175., 250., 250., 250., 250., 250., 250./)

   ! Single instance of an icebergs type initialised in icebergs_init and updated in icebergs_run
   REAL(wp), ALLOCATABLE, PUBLIC, SAVE, DIMENSION(:,:)     ::   p_calving, p_calving_hflx
   INTEGER              , PUBLIC, SAVE                     ::   numicb 
   INTEGER              , PUBLIC, SAVE, DIMENSION(nkounts) ::   kount_bergs
   INTEGER              , PUBLIC, SAVE                     ::   icbdi, icbei, icbdj, icbej
   REAL(wp)             , PUBLIC, SAVE                     ::   icb_left, icb_right
   INTEGER              , PUBLIC, SAVE                     ::   icbpack
   INTEGER              , PUBLIC, SAVE                     ::   ktberg, knberg
   INTEGER, ALLOCATABLE , PUBLIC, SAVE, DIMENSION(:)       ::   icbfldpts
   INTEGER, ALLOCATABLE , PUBLIC, SAVE, DIMENSION(:)       ::   icbflddest
   INTEGER, ALLOCATABLE , PUBLIC, SAVE, DIMENSION(:)       ::   icbfldproc

   TYPE(FLD), ALLOCATABLE, PUBLIC     , DIMENSION(:)       ::   sf_icb       ! structure: file information, fields read

END MODULE icb_oce
