MODULE icbdia

   !!======================================================================
   !!                       ***  MODULE  icbdia  ***
   !! Ocean physics:  initialise variables for iceberg budgets and diagnostics
   !!======================================================================
   !! History : 3.3.1 !  2010-01  (Martin&Adcroft) Original code
   !!            -    !  2011-03  (Madec)          Part conversion to NEMO form
   !!            -    !                            Removal of mapping from another grid
   !!            -    !  2011-04  (Alderson)       Split into separate modules
   !!            -    !  2011-05  (Alderson)       Budgets are now all here with lots
   !!            -    !                            of silly routines to call to get values in
   !!            -    !                            from the right points in the code
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !! icb_budget_end  : end        iceberg budgeting
   !! icb_budget_init : initialise iceberg budgeting
   !!----------------------------------------------------------------------
   USE par_oce        ! nemo parameters
   USE dom_oce        ! ocean domain
   USE in_out_manager ! nemo IO
   USE lib_mpp
   USE iom

   USE icb_oce        ! define iceberg arrays
   USE icbutl         ! iceberg utility routines

   IMPLICIT NONE
   PRIVATE

   REAL(wp), DIMENSION(:,:)  , POINTER, PUBLIC  :: berg_melt=>NULL()    ! Melting+erosion rate of icebergs (kg/s/m^2)
   REAL(wp), DIMENSION(:,:)  , POINTER, PUBLIC  :: melt_buoy=>NULL()    ! Buoyancy component of melting rate (kg/s/m^2)
   REAL(wp), DIMENSION(:,:)  , POINTER, PUBLIC  :: melt_eros=>NULL()    ! Erosion component of melting rate (kg/s/m^2)
   REAL(wp), DIMENSION(:,:)  , POINTER, PUBLIC  :: melt_conv=>NULL()    ! Convective component of melting rate (kg/s/m^2)
   REAL(wp), DIMENSION(:,:)  , POINTER, PUBLIC  :: bits_src=>NULL()    ! Mass flux from berg erosion into bergy bits (kg/s/m^2)
   REAL(wp), DIMENSION(:,:)  , POINTER, PUBLIC  :: bits_melt=>NULL()   ! Melting rate of bergy bits (kg/s/m^2)
   REAL(wp), DIMENSION(:,:)  , POINTER, PUBLIC  :: bits_mass=>NULL()   ! Mass distribution of bergy bits (kg/s/m^2)
   REAL(wp), DIMENSION(:,:)  , POINTER, PUBLIC  :: virtual_area=>NULL() ! Virtual surface coverage by icebergs (m^2)
   REAL(wp), DIMENSION(:,:)  , POINTER, PUBLIC  :: berg_mass=>NULL()         ! Mass distribution (kg/m^2)
   REAL(wp), DIMENSION(:,:,:), POINTER, PUBLIC  :: real_calving=>NULL() ! Calving rate into iceberg class at calving locations (kg/s)
   REAL(wp), DIMENSION(:,:)  , POINTER, PRIVATE :: tmpc=>NULL()         ! Temporary work space
   REAL(wp), DIMENSION(:)    , POINTER, PRIVATE :: zsumbuf=>NULL()         ! Temporary work space
   INTEGER , DIMENSION(:)    , POINTER, PRIVATE :: isumbuf=>NULL()         ! Temporary work space

   REAL(wp)                           , PRIVATE ::  net_berg_melt
   REAL(wp)                           , PRIVATE ::  net_bits_src
   REAL(wp)                           , PRIVATE ::  net_bits_melt
   REAL(wp)                           , PRIVATE ::  bits_mass_start, bits_mass_end
   REAL(wp)                           , PRIVATE ::  floating_heat_start, floating_heat_end
   REAL(wp)                           , PRIVATE ::  floating_mass_start, floating_mass_end
   REAL(wp)                           , PRIVATE ::  icebergs_mass_start, icebergs_mass_end
   REAL(wp)                           , PRIVATE ::  stored_start, stored_heat_start
   REAL(wp)                           , PRIVATE ::  stored_end  , stored_heat_end
   REAL(wp)                           , PRIVATE ::  net_incoming_calving, net_outgoing_calving
   REAL(wp)                           , PRIVATE ::  net_incoming_calving_heat, net_outgoing_calving_heat
   REAL(wp)                           , PRIVATE ::  net_incoming_calving_heat_used
   REAL(wp)                           , PRIVATE ::  net_calving_received, net_calving_returned, net_calving_used
   REAL(wp)                           , PRIVATE ::  net_heat_to_bergs, net_heat_to_ocean, net_melt
   REAL(wp)                           , PRIVATE ::  net_calving_to_bergs

   INTEGER                            , PRIVATE ::  nbergs_start, nbergs_end, nbergs_calved
   INTEGER                            , PRIVATE ::  nbergs_melted
   INTEGER                            , PRIVATE ::  nspeeding_tickets
   INTEGER , DIMENSION(nclasses)      , PRIVATE ::  nbergs_calved_by_class

   PUBLIC   icb_budget_end  ! routine called in xxx.F90 module
   PUBLIC   icb_budget_init ! routine called in xxx.F90 module
   PUBLIC   icb_budget      ! routine called in xxx.F90 module
   PUBLIC   icb_budget_step ! routine called in xxx.F90 module
   PUBLIC   icb_budget_put  ! routine called in xxx.F90 module
   PUBLIC   melt_budget     ! routine called in xxx.F90 module
   PUBLIC   size_budget     ! routine called in xxx.F90 module
   PUBLIC   speed_budget     ! routine called in xxx.F90 module
   PUBLIC   calving_budget  ! routine called in xxx.F90 module
   PUBLIC   incoming_budget   ! routine called in xxx.F90 module

CONTAINS

   !!-------------------------------------------------------------------------

   SUBROUTINE icb_budget_end

      IF( .NOT. ln_bergdia ) RETURN
      DEALLOCATE( berg_melt )
      DEALLOCATE( melt_buoy )
      DEALLOCATE( melt_eros )
      DEALLOCATE( melt_conv )
      DEALLOCATE( bits_src )
      DEALLOCATE( bits_melt )
      DEALLOCATE( bits_mass )
      DEALLOCATE( virtual_area )
      DEALLOCATE( berg_mass )
      DEALLOCATE( real_calving )
      DEALLOCATE( tmpc )
      IF( lk_mpp ) THEN
         DEALLOCATE( zsumbuf )
         DEALLOCATE( isumbuf )
      ENDIF

   END SUBROUTINE icb_budget_end

   !!-------------------------------------------------------------------------

   SUBROUTINE icb_budget_init( )

      IF( .NOT. ln_bergdia ) RETURN
      ALLOCATE( berg_melt    (jpi,jpj)   )           ;   berg_melt    (:,:) = 0._wp
      ALLOCATE( melt_buoy    (jpi,jpj)   )           ;   melt_buoy    (:,:) = 0._wp
      ALLOCATE( melt_eros    (jpi,jpj)   )           ;   melt_eros    (:,:) = 0._wp
      ALLOCATE( melt_conv    (jpi,jpj)   )           ;   melt_conv    (:,:) = 0._wp
      ALLOCATE( bits_src    (jpi,jpj)   )           ;   bits_src    (:,:) = 0._wp
      ALLOCATE( bits_melt   (jpi,jpj)   )           ;   bits_melt   (:,:) = 0._wp
      ALLOCATE( bits_mass   (jpi,jpj)   )           ;   bits_mass   (:,:) = 0._wp
      ALLOCATE( virtual_area (jpi,jpj)   )           ;   virtual_area (:,:) = 0._wp
      ALLOCATE( berg_mass    (jpi,jpj)   )           ;   berg_mass    (:,:) = 0._wp
      ALLOCATE( real_calving (jpi,jpj,nclasses) )    ;   real_calving (:,:,:)=0.
      ALLOCATE( tmpc(jpi,jpj) )                      ;   tmpc(:,:)=0.

      nbergs_start                   = 0
      nbergs_end                     = 0
      stored_end                     = 0._wp
      nbergs_start                   = 0._wp
      stored_start                   = 0._wp
      nbergs_melted                  = 0
      nbergs_calved                  = 0
      nbergs_calved_by_class(:)      = 0
      nspeeding_tickets              = 0
      stored_heat_end                = 0._wp
      floating_heat_end              = 0._wp
      floating_mass_end              = 0._wp
      icebergs_mass_end              = 0._wp
      bits_mass_end                 = 0._wp
      stored_heat_start              = 0._wp
      floating_heat_start            = 0._wp
      floating_mass_start            = 0._wp
      icebergs_mass_start            = 0._wp
      bits_mass_start               = 0._wp
      bits_mass_end                 = 0._wp
      net_calving_used               = 0._wp
      net_calving_to_bergs           = 0._wp
      net_heat_to_bergs              = 0._wp
      net_heat_to_ocean              = 0._wp
      net_calving_received           = 0._wp
      net_calving_returned           = 0._wp
      net_incoming_calving           = 0._wp
      net_outgoing_calving           = 0._wp
      net_incoming_calving_heat      = 0._wp
      net_incoming_calving_heat_used = 0._wp
      net_outgoing_calving_heat      = 0._wp
      net_melt                       = 0._wp
      net_berg_melt                  = 0._wp
      net_bits_melt                 = 0._wp
      net_bits_src                  = 0._wp

      floating_mass_start            = sum_mass( first_berg )
      icebergs_mass_start            = sum_mass( first_berg, justbergs=.true. )
      bits_mass_start               = sum_mass( first_berg, justbits=.true. )
      IF( lk_mpp ) THEN
         ALLOCATE( zsumbuf(23) )          ; zsumbuf(:) = 0._wp
         ALLOCATE( isumbuf(4+nclasses) )  ; isumbuf(:) = 0
         zsumbuf(1) = floating_mass_start
         zsumbuf(2) = icebergs_mass_start
         zsumbuf(3) = bits_mass_start
         CALL mpp_sum( zsumbuf(1:3), 3 )
         floating_mass_start = zsumbuf(1)
         icebergs_mass_start = zsumbuf(2)
         bits_mass_start = zsumbuf(3)
      ENDIF

   END SUBROUTINE icb_budget_init

   !!-------------------------------------------------------------------------

   SUBROUTINE icb_budget( lbudge )
      ! Arguments
      LOGICAL                         ::   lbudge
      ! Local variables
      INTEGER                         ::   k
      REAL(wp)                        ::   unused_calving, tmpsum, grdd_berg_mass, grdd_bits_mass

      IF( .NOT. ln_bergdia ) RETURN

      unused_calving            = SUM( berg_grid%calving(:,:) )
      tmpsum                    = SUM( berg_grid%floating_melt(:,:) * e1e2t(:,:) * tmask_i(:,:) )
      net_melt                  = net_melt + tmpsum * berg_dt
      net_outgoing_calving      = net_outgoing_calving + ( unused_calving + tmpsum ) * berg_dt
      tmpsum                    = SUM( berg_melt(:,:) * e1e2t(:,:) * tmask_i(:,:) )
      net_berg_melt             = net_berg_melt + tmpsum * berg_dt
      tmpsum                    = SUM( bits_src(:,:) * e1e2t(:,:) * tmask_i(:,:) )
      net_bits_src             = net_bits_src + tmpsum * berg_dt
      tmpsum                    = SUM( bits_melt(:,:) * e1e2t(:,:) * tmask_i(:,:) )
      net_bits_melt            = net_bits_melt + tmpsum * berg_dt
      tmpsum                    = SUM( p_calving(:,:) * tmask_i(:,:) )
      net_calving_returned      = net_calving_returned+tmpsum * berg_dt
      tmpsum                    = SUM( berg_grid%calving_hflx(:,:) * e1e2t(:,:) * tmask_i(:,:) )
      net_outgoing_calving_heat = net_outgoing_calving_heat + tmpsum * berg_dt   ! Units of J

      IF( lbudge ) THEN
         stored_end             = SUM( berg_grid%stored_ice(:,:,:) )
         stored_heat_end        = SUM( berg_grid%stored_heat(:,:) )
         floating_mass_end      = sum_mass( first_berg )
         icebergs_mass_end      = sum_mass( first_berg,justbergs=.true. )
         bits_mass_end         = sum_mass( first_berg,justbits=.true. )
         floating_heat_end      = sum_heat( first_berg )

         nbergs_end             = count_bergs()
         grdd_berg_mass         = SUM( berg_mass (:,:)*e1e2t(:,:)*tmask_i(:,:) )
         grdd_bits_mass        = SUM( bits_mass(:,:)*e1e2t(:,:)*tmask_i(:,:) )

         IF( lk_mpp ) THEN
            zsumbuf( 1) = stored_end
            zsumbuf( 2) = stored_heat_end
            zsumbuf( 3) = floating_mass_end
            zsumbuf( 4) = icebergs_mass_end
            zsumbuf( 5) = bits_mass_end
            zsumbuf( 6) = floating_heat_end
            zsumbuf( 7) = net_calving_returned
            zsumbuf( 8) = net_outgoing_calving
            zsumbuf( 9) = net_calving_received
            zsumbuf(10) = net_incoming_calving
            zsumbuf(11) = net_incoming_calving_heat
            zsumbuf(12) = net_incoming_calving_heat_used
            zsumbuf(13) = net_outgoing_calving_heat
            zsumbuf(14) = net_calving_used
            zsumbuf(15) = net_calving_to_bergs
            zsumbuf(16) = net_heat_to_bergs
            zsumbuf(17) = net_heat_to_ocean
            zsumbuf(18) = net_melt
            zsumbuf(19) = net_berg_melt
            zsumbuf(20) = net_bits_src
            zsumbuf(21) = net_bits_melt
            zsumbuf(22) = grdd_berg_mass
            zsumbuf(23) = grdd_bits_mass

            CALL mpp_sum( zsumbuf(1:23), 23)

            stored_end                     = zsumbuf( 1)
            stored_heat_end                = zsumbuf( 2)
            floating_mass_end              = zsumbuf( 3)
            icebergs_mass_end              = zsumbuf( 4)
            bits_mass_end                 = zsumbuf( 5)
            floating_heat_end              = zsumbuf( 6)
            net_calving_returned           = zsumbuf( 7)
            net_outgoing_calving           = zsumbuf( 8)
            net_calving_received           = zsumbuf( 9)
            net_incoming_calving           = zsumbuf(10)
            net_incoming_calving_heat      = zsumbuf(11)
            net_incoming_calving_heat_used = zsumbuf(12)
            net_outgoing_calving_heat      = zsumbuf(13)
            net_calving_used               = zsumbuf(14)
            net_calving_to_bergs           = zsumbuf(15)
            net_heat_to_bergs              = zsumbuf(16)
            net_heat_to_ocean              = zsumbuf(17)
            net_melt                       = zsumbuf(18)
            net_berg_melt                  = zsumbuf(19)
            net_bits_src                  = zsumbuf(20)
            net_bits_melt                 = zsumbuf(21)
            grdd_berg_mass                 = zsumbuf(22)
            grdd_bits_mass                = zsumbuf(23)

            isumbuf(1) = nbergs_end
            isumbuf(2) = nbergs_calved
            isumbuf(3) = nbergs_melted
            isumbuf(4) = nspeeding_tickets
            DO k = 1,nclasses
               isumbuf(4+k) = nbergs_calved_by_class(k)
            ENDDO

            CALL mpp_sum( isumbuf(1:nclasses+4), nclasses+4 )

            nbergs_end        = isumbuf(1)
            nbergs_calved     = isumbuf(2)
            nbergs_melted     = isumbuf(3)
            nspeeding_tickets = isumbuf(4)
            DO k = 1,nclasses
               nbergs_calved_by_class(k)= isumbuf(4+k)
            ENDDO

         ENDIF

         CALL report_state('stored ice','kg','',stored_start,'',stored_end,'')
         CALL report_state('floating','kg','',floating_mass_start,'',floating_mass_end,'',nbergs_end)
         CALL report_state('icebergs','kg','',icebergs_mass_start,'',icebergs_mass_end,'')
         CALL report_state('bits','kg','',bits_mass_start,'',bits_mass_end,'')
         CALL report_istate('berg #','',nbergs_start,'',nbergs_end,'')
         CALL report_ibudget('berg #','calved',nbergs_calved, &
                                      'melted',nbergs_melted, &
                                      '#',nbergs_start,nbergs_end)
         CALL report_budget('stored mass','kg','calving used',net_calving_used, &
                                               'bergs',net_calving_to_bergs, &
                                               'stored mass',stored_start,stored_end)
         CALL report_budget('floating mass','kg','calving used',net_calving_to_bergs, &
                                                 'bergs',net_melt, &
                                                 'stored mass',floating_mass_start,floating_mass_end)
         CALL report_budget('berg mass','kg','calving',net_calving_to_bergs, &
                                             'melt+eros',net_berg_melt, &
                                             'berg mass',icebergs_mass_start,icebergs_mass_end)
         CALL report_budget('bits mass','kg','eros used',net_bits_src, &
                                             'bergs',net_bits_melt, &
                                             'stored mass',bits_mass_start,bits_mass_end)
         CALL report_budget('net mass','kg','recvd',net_calving_received, &
                                            'rtrnd',net_calving_returned, &
                                            'net mass',stored_start+floating_mass_start, &
                                                       stored_end+floating_mass_end)
         CALL report_consistant('iceberg mass','kg','gridded',grdd_berg_mass,'bergs',icebergs_mass_end)
         CALL report_consistant('bits mass','kg','gridded',grdd_bits_mass,'bits',bits_mass_end)
         CALL report_state('net heat','J','',stored_heat_start+floating_heat_start,'', &
                                             stored_heat_end+floating_heat_end,'')
         CALL report_state('stored heat','J','',stored_heat_start,'',stored_heat_end,'')
         CALL report_state('floating heat','J','',floating_heat_start,'',floating_heat_end,'')
         CALL report_budget('net heat','J','net heat',net_incoming_calving_heat, &
                                           'net heat',net_outgoing_calving_heat, &
                                           'net heat',stored_heat_start+floating_heat_start, &
                                                      stored_heat_end+floating_heat_end)
         CALL report_budget('stored heat','J','calving used',net_incoming_calving_heat_used, &
                                              'bergs',net_heat_to_bergs, &
                                              'net heat',stored_heat_start,stored_heat_end)
         CALL report_budget('flting heat','J','calved',net_heat_to_bergs, &
                                              'melt',net_heat_to_ocean, &
                                              'net heat',floating_heat_start,floating_heat_end)
         IF (nn_verbose_level >= 1) THEN
            CALL report_consistant('top interface','kg','from SIS',net_incoming_calving, &
                                   'received',net_calving_received)
            CALL report_consistant('bot interface','kg','sent',net_outgoing_calving, &
                                   'returned',net_calving_returned)
         ENDIF
         WRITE(numicb,'("calved by class = ",i6,20(",",i6))') (nbergs_calved_by_class(k),k=1,nclasses)
         IF ( nspeeding_tickets .GT. 0 ) WRITE(numicb,'("speeding tickets issued = ",i6)') nspeeding_tickets

         nbergs_start                   = nbergs_end
         stored_start                   = stored_end
         nbergs_melted                  = 0
         nbergs_calved                  = 0
         nbergs_calved_by_class(:)      = 0
         nspeeding_tickets              = 0
         stored_heat_start              = stored_heat_end
         floating_heat_start            = floating_heat_end
         floating_mass_start            = floating_mass_end
         icebergs_mass_start            = icebergs_mass_end
         bits_mass_start               = bits_mass_end
         net_calving_used               = 0._wp
         net_calving_to_bergs           = 0._wp
         net_heat_to_bergs              = 0._wp
         net_heat_to_ocean              = 0._wp
         net_calving_received           = 0._wp
         net_calving_returned           = 0._wp
         net_incoming_calving           = 0._wp
         net_outgoing_calving           = 0._wp
         net_incoming_calving_heat      = 0._wp
         net_incoming_calving_heat_used = 0._wp
         net_outgoing_calving_heat      = 0._wp
         net_melt                       = 0._wp
         net_berg_melt                  = 0._wp
         net_bits_melt                 = 0._wp
         net_bits_src                  = 0._wp
      ENDIF

   END SUBROUTINE icb_budget

   !!-------------------------------------------------------------------------

   SUBROUTINE icb_budget_step
      !! things to reset at the beginning of each timestep
      !! this probably screws up fields going to diawri, so needs to be looked at - sga

      IF( .NOT. ln_bergdia ) RETURN
      berg_melt    (:,:)   = 0._wp
      melt_buoy    (:,:)   = 0._wp
      melt_eros    (:,:)   = 0._wp
      melt_conv    (:,:)   = 0._wp
      bits_src    (:,:)   = 0._wp
      bits_melt   (:,:)   = 0._wp
      bits_mass   (:,:)   = 0._wp
      berg_mass    (:,:)   = 0._wp
      virtual_area (:,:)   = 0._wp
      real_calving (:,:,:) = 0._wp

   END SUBROUTINE icb_budget_step

   !!-------------------------------------------------------------------------

   SUBROUTINE icb_budget_put

      IF( .NOT. ln_bergdia ) RETURN
      CALL iom_put( "berg_melt"         , berg_melt    (:,:)   )  ! 'Melt rate of icebergs'                    , 'kg/m2/s'
      CALL iom_put( "berg_melt_buoy"    , melt_buoy    (:,:)   )  ! 'Buoyancy component of iceberg melt rate'  , 'kg/m2/s'
      CALL iom_put( "berg_melt_eros"    , melt_eros    (:,:)   )  ! 'Erosion component of iceberg melt rate'   , 'kg/m2/s'
      CALL iom_put( "berg_melt_conv"    , melt_conv    (:,:)   )  ! 'Convective component of iceberg melt rate', 'kg/m2/s'
      CALL iom_put( "berg_virtual_area" , virtual_area (:,:)   )  ! 'Virtual coverage by icebergs'             , 'm2'
      CALL iom_put( "bits_src"         , bits_src    (:,:)   )  ! 'Mass source of bergy bits'                , 'kg/m2/s'
      CALL iom_put( "bits_melt"        , bits_melt   (:,:)   )  ! 'Melt rate of bergy bits'                  , 'kg/m2/s'
      CALL iom_put( "bits_mass"        , bits_mass   (:,:)   )  ! 'Bergy bit density field'                  , 'kg/m2'
      CALL iom_put( "berg_mass"         , berg_mass    (:,:)   )  ! 'Iceberg density field'                    , 'kg/m2'
      CALL iom_put( "berg_real_calving" , real_calving (:,:,:) )  ! 'Calving into iceberg class'               , 'kg/s'

   END SUBROUTINE icb_budget_put

   !!-------------------------------------------------------------------------

   SUBROUTINE calving_budget( ji, jj, jn, calved, heated )
      INTEGER   :: ji, jj, jn
      REAL(wp)  :: calved
      REAL(wp)  :: heated

      IF( .NOT. ln_bergdia ) RETURN
      real_calving(ji,jj,jn)     = real_calving(ji,jj,jn) + calved / berg_dt
      nbergs_calved              = nbergs_calved              + 1
      nbergs_calved_by_class(jn) = nbergs_calved_by_class(jn) + 1
      net_calving_to_bergs       = net_calving_to_bergs + calved
      net_heat_to_bergs          = net_heat_to_bergs    + heated

   END SUBROUTINE calving_budget

   !!-------------------------------------------------------------------------

   SUBROUTINE incoming_budget( kt,  calving_used, heat_used )
      INTEGER ,                 INTENT(in)  :: kt
      REAL(wp),                 INTENT(in)  :: calving_used
      REAL(wp), DIMENSION(:,:), INTENT(in)  :: heat_used

      IF( .NOT. ln_bergdia ) RETURN

      IF( kt == nit000 ) THEN
         stored_start = SUM( berg_grid%stored_ice(:,:,:) )
         IF( lk_mpp ) CALL mpp_sum( stored_start )
         WRITE(numicb,'(a,es13.6,a)')   'accumulate_calving: initial stored mass=',stored_start,' kg'

         stored_heat_start = SUM( berg_grid%stored_heat(:,:) )
         IF( lk_mpp ) CALL mpp_sum( stored_heat_start )
         WRITE(numicb,'(a,es13.6,a)')    'accumulate_calving: initial stored heat=',stored_heat_start,' J'
      ENDIF

      net_calving_received = net_calving_received + SUM( berg_grid%calving(:,:) ) * berg_dt
      net_incoming_calving = net_calving_received
      net_incoming_calving_heat = net_incoming_calving_heat +  &
                                  SUM( berg_grid%calving_hflx(:,:) * e1e2t(:,:) ) * berg_dt   ! Units of J
      net_calving_used = net_calving_used + calving_used * berg_dt
      net_incoming_calving_heat_used = net_incoming_calving_heat_used + SUM( heat_used(:,:) )

   END SUBROUTINE incoming_budget

   !!-------------------------------------------------------------------------

   SUBROUTINE size_budget(ji, jj, Wn, Ln, Abits, mass_scale, Mnew, nMbits, z1_e1e2)
      INTEGER             :: ji, jj
      REAL(wp)            :: Wn, Ln, Abits, mass_scale, Mnew, nMbits, z1_e1e2

      IF( .NOT. ln_bergdia ) RETURN
      virtual_area(ji,jj) = virtual_area(ji,jj)+(Wn*Ln+Abits)*mass_scale        ! m^2
      berg_mass(ji,jj)    = berg_mass(ji,jj) + Mnew * z1_e1e2                   ! kg/m2
      bits_mass(ji,jj)   = bits_mass(ji,jj) + nMbits * z1_e1e2                ! kg/m2

   END SUBROUTINE size_budget

   !!-------------------------------------------------------------------------

   SUBROUTINE speed_budget()

      IF( .NOT. ln_bergdia ) RETURN
      nspeeding_tickets = nspeeding_tickets + 1

   END SUBROUTINE speed_budget

   !!-------------------------------------------------------------------------

   SUBROUTINE melt_budget(ji, jj, mnew, heat, mass_scale, dM, dMbitsE, dMbitsM, dMb, dMe, dMv, z1_dt_e1e2)

      INTEGER               ::  ji, jj
      REAL(wp), INTENT(in)  ::  mnew, heat, mass_scale
      REAL(wp), INTENT(in)  ::  dM, dMbitsE, dMbitsM, dMb, dMe, dMv, z1_dt_e1e2

      IF( .NOT. ln_bergdia ) RETURN

      berg_melt    (ji,jj) = berg_melt    (ji,jj) + dM      * z1_dt_e1e2   ! kg/m2/s
      bits_src    (ji,jj) = bits_src    (ji,jj) + dMbitsE * z1_dt_e1e2   ! mass flux into bergy bitskg/m2/s
      bits_melt   (ji,jj) = bits_melt   (ji,jj) + dMbitsM * z1_dt_e1e2   ! melt rate of bergy bits kg/m2/s
      melt_buoy    (ji,jj) = melt_buoy    (ji,jj) + dMb     * z1_dt_e1e2   ! kg/m2/s
      melt_eros    (ji,jj) = melt_eros    (ji,jj) + dMe     * z1_dt_e1e2   ! erosion rate kg/m2/s
      melt_conv    (ji,jj) = melt_conv    (ji,jj) + dMv     * z1_dt_e1e2   ! kg/m2/s
      net_heat_to_ocean = net_heat_to_ocean + heat * mass_scale * berg_dt         ! J
      IF( mnew <= 0._wp ) nbergs_melted = nbergs_melted + 1                        ! Delete the berg if completely melted


   END SUBROUTINE melt_budget

   !!-------------------------------------------------------------------------

   SUBROUTINE report_state(budgetstr,budgetunits,startstr,startval,endstr,endval,delstr,nbergs)
      ! Arguments
      CHARACTER*(*), INTENT(in)           :: budgetstr, budgetunits, startstr, endstr, delstr
      REAL(wp),      INTENT(in)           :: startval, endval
      INTEGER,       INTENT(in), OPTIONAL :: nbergs

      if (present(nbergs)) then
        WRITE(numicb,100) budgetstr//' state:', &
                            startstr//' start',startval,budgetunits, &
                            endstr//' end',endval,budgetunits, &
                            'Delta '//delstr,endval-startval,budgetunits, &
                            '# of bergs',nbergs
      else
        WRITE(numicb,100) budgetstr//' state:', &
                            startstr//' start',startval,budgetunits, &
                            endstr//' end',endval,budgetunits, &
                            delstr//'Delta',endval-startval,budgetunits
      endif
      100 FORMAT(a19,3(a18,"=",es14.7,x,a2,:,","),a12,i8)
   END SUBROUTINE report_state

   !!-------------------------------------------------------------------------

   SUBROUTINE report_consistant(budgetstr,budgetunits,startstr,startval,endstr,endval)
      ! Arguments
      CHARACTER*(*), INTENT(in) :: budgetstr, budgetunits, startstr, endstr
      REAL(wp),      INTENT(in) :: startval, endval

      WRITE(numicb,200) budgetstr//' check:', &
                        startstr,startval,budgetunits, &
                        endstr,endval,budgetunits, &
                        'error',(endval-startval)/((endval+startval)+1e-30),'nd'
      200 FORMAT(a19,10(a18,"=",es14.7,x,a2,:,","))
   END SUBROUTINE report_consistant

   !!-------------------------------------------------------------------------

   SUBROUTINE report_budget(budgetstr,budgetunits,instr,inval,outstr,outval,delstr,startval,endval)
      ! Arguments
      CHARACTER*(*), INTENT(in) :: budgetstr, budgetunits, instr, outstr, delstr
      REAL(wp),      INTENT(in) :: inval, outval, startval, endval

      WRITE(numicb,200) budgetstr//' budget:', &
                        instr//' in',inval,budgetunits, &
                        outstr//' out',outval,budgetunits, &
                        'Delta '//delstr,inval-outval,budgetunits, &
                        'error',((endval-startval)-(inval-outval))/max(1.e-30,max(abs(endval-startval),abs(inval-outval))),'nd'
      200 FORMAT(a19,3(a18,"=",es14.7,x,a2,:,","),a8,"=",es10.3,x,a2)
   END SUBROUTINE report_budget

   !!-------------------------------------------------------------------------

   SUBROUTINE report_istate(budgetstr,startstr,startval,endstr,endval,delstr)
      ! Arguments
      CHARACTER*(*), INTENT(in) :: budgetstr, startstr, endstr, delstr
      INTEGER,       INTENT(in) :: startval, endval

      WRITE(numicb,100) budgetstr//' state:', &
                        startstr//' start',startval, &
                        endstr//' end',endval, &
                        delstr//'Delta',endval-startval
      100 FORMAT(a19,3(a18,"=",i14,x,:,","))
   END SUBROUTINE report_istate

   !!-------------------------------------------------------------------------

   SUBROUTINE report_ibudget(budgetstr,instr,inval,outstr,outval,delstr,startval,endval)
      ! Arguments
      CHARACTER*(*), INTENT(in) :: budgetstr, instr, outstr, delstr
      INTEGER,       INTENT(in) :: inval, outval, startval, endval

      WRITE(numicb,200) budgetstr//' budget:', &
                          instr//' in',inval, &
                          outstr//' out',outval, &
                          'Delta '//delstr,inval-outval, &
                          'error',((endval-startval)-(inval-outval))
      200 FORMAT(a19,10(a18,"=",i14,x,:,","))
   END SUBROUTINE report_ibudget
   !!-------------------------------------------------------------------------

END MODULE icbdia
