MODULE icbdyn

   !!======================================================================
   !!                       ***  MODULE  icbdyn  ***
   !! Ocean physics:  time stepping routine for iceberg tracking
   !!======================================================================
   !! History : 3.3.1 !  2010-01  (Martin&Adcroft) Original code
   !!            -    !  2011-03  (Madec)          Part conversion to NEMO form
   !!            -    !                            Removal of mapping from another grid
   !!            -    !  2011-04  (Alderson)       Split into separate modules
   !!            -    !  2011-05  (Alderson)       Replace broken grounding routine
   !!            -    !                            with one of Gurvan's suggestions (just like
   !!            -    !                            the broken one)
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   icb_init      : initialise
   !!   icb_gen       : generate test icebergs
   !!   icb_nam       : read iceberg namelist
   !!----------------------------------------------------------------------
   USE par_oce        ! NEMO parameters
   USE dom_oce        ! NEMO ocean domain
   USE phycst         ! NEMO physical constants

   USE icb_oce        ! define iceberg arrays
   USE icbutl         ! iceberg utility routines
   USE icbdia         ! iceberg budget routines

   IMPLICIT NONE
   PRIVATE

   PUBLIC   evolve_icebergs  ! routine called in xxx.F90 module

CONTAINS

   SUBROUTINE evolve_icebergs()
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE evolve_icebergs  ***
      !!
      !! ** Purpose :   iceberg evolution.
      !!
      !! ** Method  : - blah blah
      !!----------------------------------------------------------------------
      !
      REAL(wp)                        ::   uvel1 , vvel1 , u1, v1, ax1, ay1, xi1 , yj1
      REAL(wp)                        ::   uvel2 , vvel2 , u2, v2, ax2, ay2, xi2 , yj2
      REAL(wp)                        ::   uvel3 , vvel3 , u3, v3, ax3, ay3, xi3 , yj3
      REAL(wp)                        ::   uvel4 , vvel4 , u4, v4, ax4, ay4, xi4 , yj4
      REAL(wp)                        ::   uvel_n, vvel_n                  , xi_n, yj_n
      REAL(wp)                        ::   zdt, zdt_2, zdt_6, e1, e2
      LOGICAL                         ::   bounced
      TYPE(iceberg), POINTER          ::   berg
      TYPE(point)  , POINTER          ::   pt
      !!----------------------------------------------------------------------

      ! 4th order Runge-Kutta to solve:
      !    d/dt X = V,  d/dt V = A
      ! with I.C.'s:
      !    X=X1 and V=V1
      !
      !                                    ; A1=A(X1,V1)
      !  X2 = X1+dt/2*V1 ; V2 = V1+dt/2*A1 ; A2=A(X2,V2)
      !  X3 = X1+dt/2*V2 ; V3 = V1+dt/2*A2 ; A3=A(X3,V3)
      !  X4 = X1+  dt*V3 ; V4 = V1+  dt*A3 ; A4=A(X4,V4)
      !
      !  Xn = X1+dt*(V1+2*V2+2*V3+V4)/6
      !  Vn = V1+dt*(A1+2*A2+2*A3+A4)/6

      ! time steps
      zdt   = berg_dt
      zdt_2 = zdt * 0.5_wp
      zdt_6 = zdt / 6._wp

      berg => first_berg                    ! start from the first berg
      !
      DO WHILE ( ASSOCIATED(berg) )          !==  loop over all bergs  ==!
         !
         pt => berg%current_point

         bounced = .false.


         ! STEP 1 !
         ! ====== !
         xi1 = pt%xi   ;   uvel1 = pt%uvel     !**   X1 in (i,j)  ;  V1 in m/s
         yj1 = pt%yj   ;   vvel1 = pt%vvel


         !                                         !**   A1 = A(X1,V1)
         CALL accel(        xi1, e1, uvel1, uvel1, ax1,     &
            &        berg , yj1, e2, vvel1, vvel1, ay1, zdt_2 )
         !
         u1 = uvel1 / e1                           !**   V1 in d(i,j)/dt
         v1 = vvel1 / e2

         ! STEP 2 !
         ! ====== !
         !                                         !**   X2 = X1+dt/2*V1   ;   V2 = V1+dt/2*A1
         ! position using di/dt & djdt   !   V2  in m/s
         xi2 = xi1 + zdt_2 * u1          ;   uvel2 = uvel1 + zdt_2 * ax1
         yj2 = yj1 + zdt_2 * v1          ;   vvel2 = vvel1 + zdt_2 * ay1
         !
         CALL adjust_to_ground( xi2, xi1, u1, yj2, yj1, v1, bounced )

         !                                         !**   A2 = A(X2,V2)
         CALL accel(        xi2, e1, uvel2, uvel1, ax2,    &
            &        berg , yj2, e2, vvel2, vvel1, ay2, zdt_2 )
         !
         u2 = uvel2 / e1                           !**   V2 in d(i,j)/dt
         v2 = vvel2 / e2
         !
         ! STEP 3 !
         ! ====== !
         !                                         !**  X3 = X1+dt/2*V2  ;   V3 = V1+dt/2*A2; A3=A(X3)
         xi3  = xi1  + zdt_2 * u2   ;   uvel3 = uvel1 + zdt_2 * ax2
         yj3  = yj1  + zdt_2 * v2   ;   vvel3 = vvel1 + zdt_2 * ay2
         !
         CALL adjust_to_ground( xi3, xi1, u3, yj3, yj1, v3, bounced )

         !                                         !**   A3 = A(X3,V3)
         CALL accel(        xi3, e1, uvel3, uvel1, ax3,    &
            &        berg , yj3, e2, vvel3, vvel1, ay3, zdt )
         !
         u3 = uvel3 / e1                           !**   V3 in d(i,j)/dt
         v3 = vvel3 / e2

         ! STEP 4 !
         ! ====== !
         !                                         !**   X4 = X1+dt*V3   ;   V4 = V1+dt*A3
         xi4 = xi1 + zdt * u3   ;   uvel4 = uvel1 + zdt * ax3
         yj4 = yj1 + zdt * v3   ;   vvel4 = vvel1 + zdt * ay3

         CALL adjust_to_ground( xi4, xi1, u4, yj4, yj1, v4, bounced )

         !                                         !**   A4 = A(X4,V4)
         CALL accel(        xi4, e1, uvel4, uvel1, ax4,    &
            &        berg , yj4, e2, vvel4, vvel1, ay4, zdt )

         u4 = uvel4 / e1                           !**   V4 in d(i,j)/dt
         v4 = vvel4 / e2

         ! FINAL STEP !
         ! ========== !
         !                                         !**   Xn = X1+dt*(V1+2*V2+2*V3+V4)/6
         !                                         !**   Vn = V1+dt*(A1+2*A2+2*A3+A4)/6
         xi_n   = pt%xi   + zdt_6 * (  u1  + 2.*(u2  + u3 ) + u4  )
         yj_n   = pt%yj   + zdt_6 * (  v1  + 2.*(v2  + v3 ) + v4  )
         uvel_n = pt%uvel + zdt_6 * (  ax1 + 2.*(ax2 + ax3) + ax4 )
         vvel_n = pt%vvel + zdt_6 * (  ay1 + 2.*(ay2 + ay3) + ay4 )

         CALL adjust_to_ground( xi_n, xi1, uvel_n, yj_n, yj1, vvel_n, bounced )

         pt%uvel = uvel_n                        !** save in berg structure
         pt%vvel = vvel_n
         pt%xi   = xi_n
         pt%yj   = yj_n

         ! sga - update actual position
         pt%lon  = bilin_x(glamt, pt%xi, pt%yj )
         pt%lat  = bilin(gphit, pt%xi, pt%yj, 'T', 0, 0 )

         berg => berg%next                         ! switch to the next berg
         !
      END DO                                  !==  end loop over all bergs  ==!
      !
   END SUBROUTINE evolve_icebergs


   SUBROUTINE adjust_to_ground( pi, pi0, pu, pj, pj0, pv, bounced )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE adjust_to_ground  ***
      !!
      !! ** Purpose :   iceberg grounding.
      !!
      !! ** Method  : - blah blah
      !!----------------------------------------------------------------------
      REAL(wp), INTENT(inout)         ::   pi , pj    ! current iceberg position
      REAL(wp), INTENT(in   )         ::   pi0, pj0   ! previous iceberg position
      REAL(wp), INTENT(inout)         ::   pu  , pv   ! current iceberg velocities
      LOGICAL , INTENT(  out)         ::   bounced    ! bounced indicator
      !
      REAL(wp), PARAMETER ::   posn_eps = 0.05_wp     ! bouncing distance in (i,j) referential
      !
      INTEGER  ::   ii, ii0
      INTEGER  ::   ij, ij0
      INTEGER  ::   bounce_method
      !!----------------------------------------------------------------------
      !
      bounced = .false.
      bounce_method = 2
      !
      ii0 = INT( pi0+0.5 )   ;   ij0 = INT( pj0+0.5 )       ! initial gridpoint position (T-cell)
      ii  = INT( pi +0.5 )   ;   ij  = INT( pj +0.5 )       ! current     -         -
      !
      IF( ii == ii0  .AND.  ij == ij0  )   RETURN           ! berg remains in the same cell
      !
      ! map into current processor
      ii0 = ii0 - nimpp + 1
      ij0 = ij0 - njmpp + 1
      ii  = ii  - nimpp + 1
      ij  = ij  - njmpp + 1
      !
      IF(  tmask(ii,ij,1)  /=   0._wp  )   RETURN           ! berg reach a new t-cell, but an ocean one
      !
      ! From here, berg have reach land: treat grounding/bouncing
      ! -------------------------------
      bounced = .true.

      !! not obvious what should happen now
      !! if berg tries to enter a land box, the only location we can return it to is the start 
      !! position (pi0,pj0), since it has to be in a wet box to do any melting;
      !! first option is simply to set whole velocity to zero and move back to start point
      !! second option (suggested by gm) is only to set the velocity component in the (i,j) direction
      !! of travel to zero; at a coastal boundary this has the effect of sliding the berg along the coast

      SELECT CASE ( bounce_method )
         CASE ( 1 )
            pi = pi0
            pj = pj0
            pu = 0._wp
            pv = 0._wp
         CASE ( 2 )
            IF( ii0 /= ii ) THEN
               pi = pi0                   ! return back to the initial position
               pu = 0._wp                 ! zeroing of velocity in the direction of the grounding
            ENDIF
            IF( ij0 /= ij ) THEN
               pj = pj0                   ! return back to the initial position
               pv = 0._wp                 ! zeroing of velocity in the direction of the grounding
            ENDIF
      END SELECT
      !
   END SUBROUTINE adjust_to_ground


   SUBROUTINE accel(       xi, e1, uvel, uvel0, ax,                &
      &             berg , yj, e2, vvel, vvel0, ay, dt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE accel  ***
      !!
      !! ** Purpose :   compute the iceberg acceleration.
      !!
      !! ** Method  : - blah blah
      !!----------------------------------------------------------------------
      TYPE(iceberg ), POINTER           ::   berg
      REAL(wp), INTENT(in   )           ::   xi   , yj      ! berg position in (i,j) referential
      REAL(wp), INTENT(in   )           ::   uvel , vvel    ! berg velocity [m/s]
      REAL(wp), INTENT(in   )           ::   uvel0, vvel0   ! initial berg velocity [m/s]
      REAL(wp), INTENT(  out)           ::   e1, e2         ! horizontal scale factor at (xi,yj)
      REAL(wp), INTENT(inout)           ::   ax, ay         ! berg acceleration
      REAL(wp), INTENT(in   )           ::   dt             ! berg time step
      !
      REAL(wp), PARAMETER ::   alpha     = 0._wp      !
      REAL(wp), PARAMETER ::   beta      = 1._wp      !
      REAL(wp), PARAMETER ::   vel_lim   =15._wp      ! max allowed berg speed
      REAL(wp), PARAMETER ::   accel_lim = 1.e-2_wp   ! max allowed berg acceleration
      REAL(wp), PARAMETER ::   Cr0       = 0.06_wp    !
      !
      INTEGER  ::   itloop
      REAL(wp) ::   uo, vo, ui, vi, ua, va, uwave, vwave, ssh_x, ssh_y, sst, cn, hi
      REAL(wp) ::   zff, T, D, W, L, M, F
      REAL(wp) ::   drag_ocn, drag_atm, drag_ice, wave_rad
      REAL(wp) ::   c_ocn, c_atm, c_ice
      REAL(wp) ::   ampl, wmod, Cr, Lwavelength, Lcutoff, Ltop
      REAL(wp) ::   lambda, detA, A11, A12, axe, aye, D_hi
      REAL(wp) ::   uveln, vveln, us, vs, speed, loc_dx, new_speed
      !!----------------------------------------------------------------------

      ! Interpolate gridded fields to berg
      knberg = berg%number(1)
      CALL interp_flds( xi, e1, uo, ui, ua, ssh_x,                     &
         &              yj, e2, vo, vi, va, ssh_y, sst, cn, hi, zff )

      M = berg%current_point%mass
      T = berg%current_point%thickness                           ! total thickness
      D = ( rn_rho_bergs / rho_seawater ) * T   ! draught (keel depth)
      F = T - D                                    ! freeboard
      W = berg%current_point%width
      L = berg%current_point%length

      hi   = MIN( hi   , D    )
      D_hi = MAX( 0._wp, D-hi )

      ! Wave radiation
      uwave = ua - uo   ;   vwave = va - vo     ! Use wind speed rel. to ocean for wave model
      wmod  = uwave*uwave + vwave*vwave         ! The wave amplitude and length depend on the  current;
      !                                         ! wind speed relative to the ocean. Actually wmod is wmod**2 here.
      ampl        = 0.5 * 0.02025 * wmod        ! This is "a", the wave amplitude
      Lwavelength =       0.32    * wmod        ! Surface wave length fitted to data in table at
      !                                         ! http://www4.ncsu.edu/eos/users/c/ceknowle/public/chapter10/part2.html
      Lcutoff     = 0.125 * Lwavelength
      Ltop        = 0.25  * Lwavelength
      Cr          = Cr0 * MIN(  MAX( 0., (L-Lcutoff) / ((Ltop-Lcutoff)+1.e-30)) , 1.)  ! Wave radiation coefficient
      !                                         ! fitted to graph from Carrieres et al.,  POAC Drift Model.
      wave_rad    = 0.5 * rho_seawater / M * Cr * grav * ampl * MIN( ampl,F ) * (2.*W*L) / (W+L)
      wmod        = SQRT( ua*ua + va*va )       ! Wind speed
      IF( wmod /= 0._wp ) THEN
         uwave = ua/wmod ! Wave radiation force acts in wind direction ...       !!gm  this should be the wind rel. to ocean ?
         vwave = va/wmod
      ELSE
         uwave = 0.   ;    vwave=0.   ;    wave_rad=0. ! ... and only when wind is present.     !!gm  wave_rad=0. is useless
      ENDIF

      ! Weighted drag coefficients
      c_ocn = rho_seawater / M * (0.5*Cd_wv*W*(D_hi)+Cd_wh*W*L)
      c_atm = rho_air      / M * (0.5*Cd_av*W*F     +Cd_ah*W*L)
      c_ice = rho_ice      / M * (0.5*Cd_iv*W*hi              )
      IF( abs(ui) + abs(vi) == 0._wp )   c_ice = 0._wp

      uveln = uvel   ;   vveln = vvel ! Copy starting uvel, vvel
      !
      DO itloop = 1, 2  ! Iterate on drag coefficients
         !
         us = 0.5 * ( uveln + uvel )
         vs = 0.5 * ( vveln + vvel )
         drag_ocn = c_ocn * SQRT( (us-uo)*(us-uo) + (vs-vo)*(vs-vo) )
         drag_atm = c_atm * SQRT( (us-ua)*(us-ua) + (vs-va)*(vs-va) )
         drag_ice = c_ice * SQRT( (us-ui)*(us-ui) + (vs-vi)*(vs-vi) )
         !
         ! Explicit accelerations
         !axe= zff*vvel -grav*ssh_x +wave_rad*uwave &
         !    -drag_ocn*(uvel-uo) -drag_atm*(uvel-ua) -drag_ice*(uvel-ui)
         !aye=-zff*uvel -grav*ssh_y +wave_rad*vwave &
         !    -drag_ocn*(vvel-vo) -drag_atm*(vvel-va) -drag_ice*(vvel-vi)
         axe = -grav * ssh_x + wave_rad * uwave
         aye = -grav * ssh_y + wave_rad * vwave
         IF( alpha > 0._wp ) THEN   ! If implicit, use time-level (n) rather than RK4 latest
            axe = axe + zff*vvel0
            aye = aye - zff*uvel0
         else
            axe=axe+zff*vvel
            aye=aye-zff*uvel
         endif
         if (beta>0.) then ! If implicit, use time-level (n) rather than RK4 latest
            axe=axe-drag_ocn*(uvel0-uo) -drag_atm*(uvel0-ua) -drag_ice*(uvel0-ui)
            aye=aye-drag_ocn*(vvel0-vo) -drag_atm*(vvel0-va) -drag_ice*(vvel0-vi)
         else
            axe=axe-drag_ocn*(uvel-uo) -drag_atm*(uvel-ua) -drag_ice*(uvel-ui)
            aye=aye-drag_ocn*(vvel-vo) -drag_atm*(vvel-va) -drag_ice*(vvel-vi)
         endif

         ! Solve for implicit accelerations
         IF( alpha + beta > 0._wp ) THEN
            lambda = drag_ocn + drag_atm + drag_ice
            A11 = 1.+beta*dt*lambda
            A12 = alpha*dt*zff
            detA = 1._wp / ( A11*A11 + A12*A12 )
            ax = detA * ( A11*axe+A12*aye )
            ay = detA * ( A11*aye-A12*axe )
         else
            ax=axe   ;   ay=aye
         endif

         uveln = uvel0 + dt*ax
         vveln = vvel0 + dt*ay
         !
      END DO      ! itloop

      IF( rn_speed_limit > 0.) THEN       ! Limit speed of bergs based on a CFL criteria (if asked)
         speed = SQRT( uveln*uveln + vveln*vveln )    ! Speed of berg
         IF( speed > 0.) THEN
            loc_dx = MIN( e1, e2 )                          ! minimum grid spacing
            new_speed = loc_dx / dt * rn_speed_limit     ! Speed limit as a factor of dx / dt
            IF( new_speed < speed ) THEN
               uveln = uveln * ( new_speed / speed )        ! Scale velocity to reduce speed
               vveln = vveln * ( new_speed / speed )        ! without changing the direction
               CALL speed_budget()
            ENDIF
         ENDIF
      ENDIF
      !                                      ! check the speed and acceleration limits
      IF( ABS( uveln ) > vel_lim   .OR. ABS( vveln ) > vel_lim   )   &
         WRITE(numicb,'("pe=",i3,x,a)') narea,'Dump triggered by excessive velocity'
      IF( ABS( ax    ) > accel_lim .OR. ABS( ay    ) > accel_lim )   &
         WRITE(numicb,'("pe=",i3,x,a)') narea,'Dump triggered by excessive acceleration'
      !
   END SUBROUTINE accel

END MODULE icbdyn
