MODULE icbini

   !!======================================================================
   !!                       ***  MODULE  icbini  ***
   !! Ocean physics:  initialise variables for iceberg tracking
   !!======================================================================
   !! History : 3.3.1 !  2010-01  (Martin&Adcroft) Original code
   !!            -    !  2011-03  (Madec)          Part conversion to NEMO form
   !!            -    !                            Removal of mapping from another grid
   !!            -    !  2011-04  (Alderson)       Split into separate modules
   !!            -    !                            Restore restart routines
   !!            -    !  2011-05  (Alderson)       generate_test_icebergs restored
   !!            -    !  2011-05  (Alderson)       new forcing arrays with extra halo
   !!            -    !  2011-05  (Alderson)       north fold exchange arrays added
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   icb_init      : initialise
   !!   icb_gen       : generate test icebergs
   !!   icb_nam       : read iceberg namelist
   !!----------------------------------------------------------------------
   USE dom_oce        ! ocean domain
   USE in_out_manager ! IO routines and numout in particular
   USE lib_mpp        ! mpi library and lk_mpp in particular
   USE sbc_oce
   USE iom
   USE fldread
   USE lbclnk

   USE icb_oce        ! define iceberg arrays
   USE icbutl         ! iceberg utility routines
   USE icbrst         ! iceberg restart routines
   USE icbtrj         ! iceberg trajectory I/O routines
   USE icbdia         ! iceberg budget routines

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=100), PRIVATE ::   cn_dir       = './'    ! Root directory for location of icb files
   TYPE(FLD_N), PRIVATE        ::   sn_icb                 ! information about the calving file to be read

   PRIVATE  icb_alloc      ! routine called in xxx.F90 module
   PUBLIC   icb_init       ! routine called in xxx.F90 module
   PUBLIC   icb_gen        ! routine called in xxx.F90 module
   PUBLIC   icb_nam        ! routine called in xxx.F90 module

CONTAINS

   INTEGER FUNCTION icb_alloc()
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE icb_alloc  ***
      !!----------------------------------------------------------------------
      !
      INTEGER                              ::   ill
      !
      ! open ascii output file or files for iceberg status information
      ! note that we choose to do this on all processors since we cannot predict where
      ! icebergs will be ahead of time
      !
      CALL ctl_opn( numicb, 'icebergs.stat', 'REPLACE', 'FORMATTED', 'SEQUENTIAL', &
                    -1, numout, lwp, narea )

      icb_alloc = 0
      ALLOCATE(berg_grid, STAT=ill)
      icb_alloc = icb_alloc + ill
      !
      ALLOCATE( berg_grid%calving      (jpi,jpj)   , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( berg_grid%calving_hflx (jpi,jpj)   , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( berg_grid%stored_heat  (jpi,jpj)   , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( berg_grid%floating_melt(jpi,jpj)   , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( berg_grid%maxclass(jpi,jpj)        , STAT=ill)
      icb_alloc = icb_alloc + ill
      !
      ALLOCATE( berg_grid%stored_ice   (jpi,jpj,nclasses) , STAT=ill)
      icb_alloc = icb_alloc + ill
      !
      ALLOCATE( berg_grid%tmp (jpi,jpj) , STAT=ill)
      icb_alloc = icb_alloc + ill
      !
      ! expanded arrays for bilinear interpolation
      ALLOCATE( uo_e(0:jpi+1,0:jpj+1) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( vo_e(0:jpi+1,0:jpj+1) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( ff_e(0:jpi+1,0:jpj+1) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( ua_e(0:jpi+1,0:jpj+1) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( va_e(0:jpi+1,0:jpj+1) , STAT=ill)
      icb_alloc = icb_alloc + ill
#if defined key_lim2 || defined key_lim3
      ALLOCATE( ui_e(0:jpi+1,0:jpj+1) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( vi_e(0:jpi+1,0:jpj+1) , STAT=ill)
      icb_alloc = icb_alloc + ill
#endif
      ALLOCATE( ssh_e(0:jpi+1,0:jpj+1) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( initial_width    (nclasses) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( initial_length   (nclasses) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( p_calving(jpi,jpj) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( p_calving_hflx(jpi,jpj) , STAT=ill)
      icb_alloc = icb_alloc + ill

      ALLOCATE( icbfldpts(jpi) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( icbflddest(jpi) , STAT=ill)
      icb_alloc = icb_alloc + ill
      ALLOCATE( icbfldproc(jpni) , STAT=ill)
      icb_alloc = icb_alloc + ill

      IF( lk_mpp        )   CALL mpp_sum ( icb_alloc )
      IF( icb_alloc > 0 )   CALL ctl_warn('icb_alloc: allocation of arrays failed')

   END FUNCTION icb_alloc

   SUBROUTINE icb_init( dt, kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dom_init  ***
      !!
      !! ** Purpose :   iceberg initialization.
      !!
      !! ** Method  : - blah blah
      !!----------------------------------------------------------------------
      REAL(wp)                , INTENT(in) ::   dt      ! iceberg time-step (rdt*nn_fsbc)
      INTEGER                 , INTENT(in) ::   kt      ! time step number
      !
      INTEGER                              ::   ji, jj, jn, i1, j1
      INTEGER                              ::   ii, inum, ivar, k1
      INTEGER                              ::   istat1, istat2, istat3
      CHARACTER(len=300)                   ::   sdist
      !!----------------------------------------------------------------------

      CALL icb_nam               ! Read and print namelist parameters
      IF( .NOT. ln_icebergs ) RETURN

      !                          ! allocate gridded fields
      IF( icb_alloc() /= 0 )   CALL ctl_stop( 'STOP', 'icb_alloc : unable to allocate arrays' )

      ! set parameters (mostly from namelist)
      !
      berg_dt                  = dt
      initial_width (:) = SQRT(  rn_initial_mass(:) / ( rn_LoW_ratio*rn_rho_bergs*rn_initial_thickness(:) )  )
      initial_length(:) = rn_LoW_ratio * initial_width(:)

      berg_grid%calving      (:,:)   = 0._wp
      berg_grid%calving_hflx (:,:)   = 0._wp
      berg_grid%stored_heat  (:,:)   = 0._wp
      berg_grid%floating_melt(:,:)   = 0._wp
      berg_grid%maxclass      (:,:)  = nclasses
      berg_grid%stored_ice   (:,:,:) = 0._wp
      berg_grid%tmp          (:,:)   = 0._wp
      p_calving        (:,:)   = 0._wp
      p_calving_hflx   (:,:)   = 0._wp

      ! domain for icebergs
      IF( lk_mpp .AND. jpni == 1 ) &
                  CALL ctl_stop('icbinit: having ONE processor in x currently does not work')

      ! for the north fold we work out which points communicate by asking
      ! lbc_lnk to pass processor number (valid even in single processor case)
      ! borrow p_calving arrays for this
      !
      ! pack i and j together using a scaling of a power of 10
      icbpack = 10000
      IF( jpiglo >= icbpack ) CALL ctl_stop('icbini: processor index packing failure')
      icbfldproc(:) = -1

      DO jj = 1, jpj
         DO ji = 1, jpi
            p_calving_hflx(ji,jj) = narea
            p_calving(ji,jj)      = icbpack*(njmpp+jj-1) + nimpp+ji-1
         ENDDO
      ENDDO
      CALL lbc_lnk( p_calving_hflx, 'T', 1._wp )
      CALL lbc_lnk( p_calving     , 'T', 1._wp )

      ! work out interior of processor from exchange array
      ! first entry with narea for this processor is left hand interior index
      ! last  entry                               is right hand interior index
      jj = jpj/2
      icbdi = -1
      icbei = -1
      DO ji = 1,jpi
         k1 = INT( p_calving(ji,jj) )
         j1 = INT( k1/icbpack )
         i1 = k1 - j1*icbpack
         IF( i1 == nimpp+ji-1 ) THEN
            IF( icbdi < 0 ) THEN
               icbdi = ji
            ELSE
               icbei = ji
            ENDIF
         ENDIF
      END DO
     
      ! repeat for j direction
      ji = jpi/2
      icbdj = -1
      icbej = -1
      DO jj = 1,jpj
         k1 = INT( p_calving(ji,jj) )
         j1 = INT( k1/icbpack )
         i1 = k1 - j1*icbpack
         IF( j1 == njmpp+jj-1 ) THEN
            IF( icbdj < 0 ) THEN
               icbdj = jj
            ELSE
               icbej = jj
            ENDIF
         ENDIF
      END DO
         
      ! special for east-west boundary exchange we save the destination index
      i1 = MAX( icbdi-1, 1)
      k1 = INT( p_calving(i1,jpj/2) )
      jj = INT( k1/icbpack )
      icb_left = REAL( k1 - icbpack*jj, wp )
      i1 = MIN( icbei+1, jpi )
      k1 = INT( p_calving(i1,jpj/2) )
      jj = INT( k1/icbpack )
      icb_right = REAL( k1 - icbpack*jj, wp )
      
      ! north fold
      IF( npolj > 0 ) THEN
         !
         ! icebergs in row icbej+1 get passed across fold
         icbfldpts(:)  = INT( p_calving(:,icbej+1) )
         icbflddest(:) = INT( p_calving_hflx(:,icbej+1) )
         !
         ! work out list of unique processors to talk to
         DO ji = icbdi, icbei
            ii = icbflddest(ji)
            DO jn = 1, jpni
               IF( icbfldproc(jn) == -1 ) THEN
                  icbfldproc(jn) = ii
                  EXIT
               ENDIF
               IF( icbfldproc(jn) == ii ) EXIT
            ENDDO
         ENDDO
      ENDIF
      !
      IF( nn_verbose_level > 0) THEN
         WRITE(numicb,*) 'processor ', narea
         WRITE(numicb,*) 'jpi, jpj   ', jpi, jpj
         WRITE(numicb,*) 'nldi, nlei ', nldi, nlei
         WRITE(numicb,*) 'nldj, nlej ', nldj, nlej
         WRITE(numicb,*) 'berg i interior ', icbdi, icbei
         WRITE(numicb,*) 'berg j interior ', icbdj, icbej
         WRITE(numicb,*) 'berg left       ', icb_left
         WRITE(numicb,*) 'berg right      ', icb_right
         jj = jpj/2
         WRITE(numicb,*) "central j line:"
         WRITE(numicb,*) "i processor"
         WRITE(numicb,*) (INT(p_calving_hflx(ji,jj)), ji=1,jpi)
         WRITE(numicb,*) "i point"
         WRITE(numicb,*) (INT(p_calving(ji,jj)), ji=1,jpi)
         ji = jpi/2
         WRITE(numicb,*) "central i line:"
         WRITE(numicb,*) "j processor"
         WRITE(numicb,*) (INT(p_calving_hflx(ji,jj)), jj=1,jpj)
         WRITE(numicb,*) "j point"
         WRITE(numicb,*) (INT(p_calving(ji,jj)), jj=1,jpj)
         IF( npolj > 0 ) THEN
            WRITE(numicb,*) 'north fold destination points '
            WRITE(numicb,*) icbfldpts
            WRITE(numicb,*) 'north fold destination procs  '
            WRITE(numicb,*) icbflddest
         ENDIF
         CALL flush(numicb)
      ENDIF
      
      p_calving(:,:)      = 0._wp
      p_calving_hflx(:,:) = 0._wp

      ! assign each new iceberg with a unique number constructed from the processor number
      ! and incremented by the total number of processors
      kount_bergs(:) = 0
      kount_bergs(1) = narea - jpnij

      ! when not generating test icebergs we need to setup calving file
      IF( nn_test_icebergs < 0 ) THEN
         !
         ! maximum distribution class array does not change in time so read it once
         sdist = TRIM( cn_dir )//TRIM( sn_icb%clname )
         CALL iom_open ( sdist, inum )                              ! open file
         ivar = iom_varid( inum, 'maxclass', ldstop=.FALSE. )
         IF( ivar > 0 ) THEN
            CALL iom_get  ( inum, jpdom_data, 'maxclass', p_calving )   ! read the max distribution array
            berg_grid%maxclass(:,:) = INT( p_calving )
            p_calving(:,:) = 0._wp
         ENDIF
         CALL iom_close( inum )                                     ! close file
         !
         WRITE(numicb,*)
         WRITE(numicb,*) '          calving read in a file'
         ALLOCATE( sf_icb(1), STAT=istat1 )         ! Create sf_icb structure (calving)
         ALLOCATE( sf_icb(1)%fnow(jpi,jpj,1), STAT=istat2 )
         ALLOCATE( sf_icb(1)%fdta(jpi,jpj,1,2), STAT=istat3 )
         IF( istat1+istat2+istat3 > 0 ) THEN
            CALL ctl_stop( 'sbc_icb: unable to allocate sf_icb structure' )   ;   RETURN
         ENDIF
         !                                          ! fill sf_icb with the namelist (sn_icb) and control print
         CALL fld_fill( sf_icb, (/ sn_icb /), cn_dir, 'icb_init', 'read calving data', 'namicb' )
         !
      ENDIF

      IF( .NOT.ln_rstart ) THEN
         IF( nn_test_icebergs > 0 ) CALL icb_gen()
      ELSE
         IF( nn_test_icebergs > 0 ) THEN
            CALL icb_gen()
         ELSE
            !
            CALL icebergs_read_restart()
            restarted_bergs = .TRUE.
         ENDIF
      ENDIF
      !
      IF( nn_sample_rate .GT. 0 ) CALL traj_init( nitend )
      !
      CALL icb_budget_init()
      !
      IF( nn_verbose_level >= 2 )   CALL print_bergs('icb_init, initial status', nit000-1)
      !
   END SUBROUTINE icb_init

   SUBROUTINE icb_gen()

      ! Local variables
      INTEGER                         ::   i, j, nbergs
      TYPE(iceberg)                   ::   localberg ! NOT a pointer but an actual local variable
      TYPE(point)                     ::   localpt
      INTEGER                         ::   iyr, imon, iday, ihr, imin, isec
      INTEGER                         ::   iberg

      ! For convenience
      iberg = nn_test_icebergs

      ! call get_date(Time, iyr, imon, iday, ihr, imin, isec)
      ! Convert nemo time variables from dom_oce into local versions
      iyr = nyear
      imon = nmonth
      iday = nday
      ihr = INT(nsec_day/3600)
      imin = INT((nsec_day-ihr*3600)/60)
      isec = nsec_day - ihr*3600 - imin*60

      ! no overlap for icebergs since we want only one instance of each across the whole domain
      ! so restrict area of interest
      ! use tmask here because tmask_i has been doctored on one side of the north fold line

      DO j = icbdj,icbej
         DO i = icbdi,icbei
            IF (tmask(i,j,1) .GT. 0._wp .AND. &
                gphit(i,j) .GT. rn_test_box(3) .AND. gphit(i,j) .LT. rn_test_box(4) .AND. &
                glamt(i,j) .GT. rn_test_box(1) .AND. glamt(i,j) .LT. rn_test_box(2)) THEN
               localberg%mass_scaling = rn_mass_scaling(iberg)
               localpt%xi = REAL( nimpp+i-1, wp )
               localpt%yj = REAL( njmpp+j-1, wp )
               localpt%lon = bilin(glamt, localpt%xi, localpt%yj, 'T', 0, 0 )
               localpt%lat = bilin(gphit, localpt%xi, localpt%yj, 'T', 0, 0 )
               localpt%mass = rn_initial_mass(iberg)
               localpt%thickness = rn_initial_thickness(iberg)
               localpt%width = initial_width(iberg)
               localpt%length = initial_length(iberg)
               localpt%year = iyr
               localpt%day = FLOAT(iday)+(FLOAT(ihr)+FLOAT(imin)/60._wp)/24._wp
               localpt%mass_of_bits = 0._wp
               localpt%heat_density = 0._wp
               localpt%uvel = 0._wp
               localpt%vvel = 0._wp
               CALL increment_kounter()
               localberg%number(:) = kount_bergs(:)
               call add_new_berg_to_list(localberg, localpt)
            ENDIF
         ENDDO
      ENDDO

      nbergs = count_bergs()
      IF( lk_mpp ) CALL mpp_sum(nbergs)
      WRITE(numicb,'(a,i6,a)') 'diamonds, icb_gen: ',nbergs,' were generated'

   END SUBROUTINE icb_gen

   SUBROUTINE icb_nam
      !!----------------------------------------------------------------------
      !!                     ***  ROUTINE icb_nam  ***
      !!
      !! ** Purpose :   read domaine namelists and print the variables.
      !!
      !! ** input   : - namberg namelist
      !!----------------------------------------------------------------------

      NAMELIST/namberg/ ln_icebergs, ln_bergdia, nn_sample_rate, rn_initial_mass,                  &
         &              rn_distribution, rn_mass_scaling, rn_initial_thickness, nn_verbose_write,  &
         &              rn_rho_bergs, rn_LoW_ratio, nn_verbose_level, ln_operator_splitting,       &
         &              rn_bits_erosion_fraction, rn_sicn_shift, ln_passive_mode,                  &
         &              ln_time_average_weight, nn_test_icebergs, rn_test_box, rn_speed_limit,     &
         &              cn_dir, sn_icb
      INTEGER   ::   k  ! loop counter
      REAL(wp)  ::   dfact
      !!----------------------------------------------------------------------

      ! (NB: frequency positive => hours, negative => months)
      !            !   file     ! frequency !  variable  ! time intep !  clim  ! 'yearly' or ! weights  ! rotation   !
      !            !   name     !  (hours)  !   name     !   (T/F)    !  (T/F) !  'monthly'  ! filename ! pairs      !
      sn_icb = FLD_N( 'calving' ,    -1     , 'calving'  ,  .TRUE.    , .TRUE. ,   'yearly'  , ''       , ''         )

      REWIND( numnam )              ! Namelist namrun : iceberg parameters
      READ  ( numnam, namberg )

      IF( .NOT. ln_icebergs .AND. lwp ) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'icb_nam : iceberg initialization through namberg namelist read'
         WRITE(numout,*) '~~~~~~~~~~~~~ '
         WRITE(numout,*) 'NO icebergs used'
         RETURN
      ENDIF

      IF( nn_test_icebergs > nclasses ) THEN
          IF( lwp ) WRITE(numout,*) 'Resetting nn_test_icebergs to ',nclasses
          nn_test_icebergs = nclasses
      ENDIF

      dfact = SUM( rn_distribution )
      IF( dfact < 1._wp ) THEN
         IF( dfact <= 0._wp ) THEN
            CALL ctl_stop('icb_init: sum of berg distribution equal to zero')
         ELSE
            rn_distribution(:) = rn_distribution(:) / dfact
            CALL ctl_warn('icb_init: sum of berg input distribution not equal to one and so RESCALED')
         ENDIF
      ENDIF

      IF(lwp) THEN                  ! control print
         WRITE(numout,*)
         WRITE(numout,*) 'icb_nam : iceberg initialization through namberg namelist read'
         WRITE(numout,*) '~~~~~~~~~~~~~ '
         WRITE(numout,*) '   Calculate budgets                                            ln_bergdia       = ', ln_bergdia
         WRITE(numout,*) '   Period between sampling of position for trajectory storage   nn_sample_rate = ', nn_sample_rate
         WRITE(numout,*) '   Mass thresholds between iceberg classes (kg)                 rn_initial_mass     ='
         DO k=1,nclasses
            WRITE(numout,'(a,f15.2)') '                                                                ',rn_initial_mass(k)
         ENDDO
         WRITE(numout,*) '   Fraction of calving to apply to this class (non-dim)         rn_distribution     ='
         DO k=1,nclasses
            WRITE(numout,'(a,f10.2)') '                                                                ',rn_distribution(k)
         ENDDO
         WRITE(numout,*) '   Ratio between effective and real iceberg mass (non-dim)      rn_mass_scaling     = '
         DO k=1,nclasses
            WRITE(numout,'(a,f10.2)') '                                                                ',rn_mass_scaling(k)
         ENDDO
         WRITE(numout,*) '   Total thickness of newly calved bergs (m)                   rn_initial_thickness = '
         DO k=1,nclasses
            WRITE(numout,'(a,f10.2)') '                                                                ',rn_initial_thickness(k)
         ENDDO
         WRITE(numout,*) '   Timesteps between verbose messages                           nn_verbose_write    = ', nn_verbose_write

         WRITE(numout,*) '   Density of icebergs                           rn_rho_bergs   = ', rn_rho_bergs
         WRITE(numout,*) '   Initial ratio L/W for newly calved icebergs   rn_LoW_ratio   = ', rn_LoW_ratio
         WRITE(numout,*) '   Turn on more verbose output                         level = ', nn_verbose_level
         WRITE(numout,*) '   Use first order operator splitting for thermodynamics    ',   &
            &                    'use_operator_splitting = ', ln_operator_splitting
         WRITE(numout,*) '   Fraction of erosion melt flux to divert to bergy bits    ',   &
            &                    'bits_erosion_fraction = ', rn_bits_erosion_fraction

         WRITE(numout,*) '   Shift of sea-ice concentration in erosion flux modulation ',   &
            &                    '(0<sicn_shift<1)  sicn_shift              = ', rn_sicn_shift
         WRITE(numout,*) '   Do not add freshwater flux from icebergs to ocean                ',   &
            &                    '                  passive_mode            = ', ln_passive_mode
         WRITE(numout,*) '   Time average the weight on the ocean   time_average_weight     = ', ln_time_average_weight
         WRITE(numout,*) '   Create icebergs in absence of a restart file   nn_test_icebergs  = ', nn_test_icebergs
         WRITE(numout,*) '                   in lon/lat box                                         = ', rn_test_box
         WRITE(numout,*) '   CFL speed limit for a berg            speed_limit             = ', rn_speed_limit
         WRITE(numout,*) '   Writing Iceberg status information to icebergs.stat file        '
      ENDIF
      !
   END SUBROUTINE icb_nam

   !!======================================================================

END MODULE icbini
