MODULE icblbc

   !!======================================================================
   !!                       ***  MODULE  icblbc  ***
   !! Ocean physics:  routines to handle boundary exchanges for icebergs
   !!======================================================================
   !! History : 3.3.1 !  2010-01  (Martin&Adcroft) Original code
   !!            -    !  2011-03  (Madec)          Part conversion to NEMO form
   !!            -    !                            Removal of mapping from another grid
   !!            -    !  2011-04  (Alderson)       Split into separate modules
   !!            -    !  2011-05  (Alderson)       MPP exchanges written based on lib_mpp
   !!            -    !  2011-05  (Alderson)       MPP and single processor boundary
   !!            -    !                            conditions added
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   mpp_send_bergs   :  In MPP pass icebergs from linked list between processors
   !!                       as they advect around
   !!                       Lagrangian processes cannot be handled by existing NEMO MPP
   !!                       routines because they do not lie on regular jpi,jpj grids
   !!                       Processor exchanges are handled as in lib_mpp whenever icebergs step 
   !!                       across boundary of interior domain (icbdi-icbei, icbdj-icbej)
   !!                       so that iceberg does not exist in more than one processor
   !!                       North fold exchanges controlled by three arrays:
   !!                          icbflddest - unique processor numbers that current one exchanges with
   !!                          icbfldproc - processor number that current grid point exchanges with
   !!                          icbfldpts  - packed i,j point in exchanging processor
   !!----------------------------------------------------------------------

   USE par_oce                             ! ocean parameters
   USE dom_oce                             ! ocean domain
   USE in_out_manager                      ! IO parameters
   USE lib_mpp                             ! MPI code and lk_mpp in particular
   USE icb_oce                             ! define iceberg arrays
   USE icbutl                              ! iceberg utility routines

   IMPLICIT NONE
   PRIVATE

#if defined key_mpp_mpi

!$AGRIF_DO_NOT_TREAT
   INCLUDE 'mpif.h'
!$AGRIF_END_DO_NOT_TREAT

   TYPE, PUBLIC :: buffer
      INTEGER :: size=0
      REAL(wp), DIMENSION(:,:), POINTER :: data
   END TYPE buffer

   TYPE(buffer), POINTER           ::   obuffer_n=>NULL() , ibuffer_n=>NULL()
   TYPE(buffer), POINTER           ::   obuffer_s=>NULL() , ibuffer_s=>NULL()
   TYPE(buffer), POINTER           ::   obuffer_e=>NULL() , ibuffer_e=>NULL()
   TYPE(buffer), POINTER           ::   obuffer_w=>NULL() , ibuffer_w=>NULL()

   ! north fold exchange buffers
   TYPE(buffer), POINTER           ::   obuffer_f=>NULL() , ibuffer_f=>NULL()

   INTEGER, PARAMETER, PUBLIC      ::   delta_buf = 25             ! Size by which to increment buffers
   INTEGER, PARAMETER, PUBLIC      ::   buffer_width = 15+nkounts  ! items to store for each berg

#endif

   PUBLIC   lbc_send_bergs
   PRIVATE  lbc_nfld_bergs
   PUBLIC   mpp_send_bergs
   PUBLIC   dealloc_buffers

#if defined key_mpp_mpi
   PRIVATE  mpp_nfld_bergs
   PRIVATE  dealloc_buffer
   PRIVATE  pack_berg_into_buffer
   PRIVATE  unpack_berg_from_buffer
   PRIVATE  increase_buffer
   PRIVATE  increase_ibuffer
#endif

   !!-------------------------------------------------------------------------
CONTAINS

   SUBROUTINE lbc_send_bergs()
      !!----------------------------------------------------------------------
      !!                 ***  SUBROUTINE lbc_send_bergs  ***
      !!
      !! ** Purpose :   in non-mpp case need to deal with cyclic conditions
      !!                including north-fold
      !!----------------------------------------------------------------------
      ! Local variables
      TYPE(iceberg), POINTER :: this
      TYPE(point)  , POINTER :: pt
      INTEGER                :: ine

      !! periodic east/west boundaries
      !! =============================

      IF( nperio == 1 .OR. nperio == 4 .OR. nperio == 6 ) THEN

         this => first_berg
         DO WHILE( ASSOCIATED(this) )
            pt => this%current_point
            ine = INT( pt%xi + 0.5 )
            IF( ine .GT. nimpp+icbei-1 ) THEN
               pt%xi = icb_right + MOD(pt%xi, 1._wp ) - 1._wp
            ELSE IF( ine .LT. nimpp+icbdi-1 ) THEN
               pt%xi = icb_left + MOD(pt%xi, 1._wp )
            ENDIF
            this => this%next
         ENDDO

      ENDIF

      !! north/south boundaries
      !! ======================

      !! south symmetric

      IF( nperio == 2 ) THEN
         CALL ctl_stop(' south symmetric condition not implemented for icebergs')
      ENDIF

      !! north fold

      IF( nperio == 3 .OR. nperio == 4 .OR. nperio == 5 .OR. nperio == 6 ) CALL lbc_nfld_bergs()

   END SUBROUTINE lbc_send_bergs

   !!-------------------------------------------------------------------------

   SUBROUTINE lbc_nfld_bergs()
      !!----------------------------------------------------------------------
      !!                 ***  SUBROUTINE lbc_nfld_bergs  ***
      !!
      !! ** Purpose :   single processor north fold exchange
      !!----------------------------------------------------------------------
      ! 
      ! Local variables
      TYPE(iceberg), POINTER :: this
      TYPE(point)  , POINTER :: pt
      INTEGER                :: ine, jne, kpts
      INTEGER                :: jiglo, jjglo

      this => first_berg
      DO WHILE( ASSOCIATED(this) )
         pt => this%current_point
         jne = INT( pt%yj + 0.5 )
         IF( jne .GT. njmpp+icbej-1 ) THEN
            !
            ine = INT( pt%xi + 0.5 )
            kpts  = icbfldpts (ine-nimpp+1)
            !
            ! moving across the cut line means both position and
            ! velocity must change
            jjglo = INT( kpts/icbpack )
            jiglo = kpts - icbpack*jjglo
            pt%xi = jiglo - ( pt%xi - REAL(ine,wp) )
            pt%yj = jjglo - ( pt%yj - REAL(jne,wp) )
            pt%uvel = -1._wp * pt%uvel
            pt%vvel = -1._wp * pt%vvel
         ENDIF
         this => this%next
      ENDDO

   END SUBROUTINE lbc_nfld_bergs

   !!-------------------------------------------------------------------------

#if defined key_mpp_mpi

   SUBROUTINE mpp_send_bergs()
      !!----------------------------------------------------------------------
      !!                 ***  SUBROUTINE mpp_send_bergs  ***
      !!
      !! ** Purpose :   multi processor exchange
      !!
      !! ** Method  :   identify direction for exchange, pack into a buffer
      !!                which is basically a real array and delete from linked list
      !!                length of buffer is exchanged first with receiving processor
      !!                then buffer is sent if necessary
      !!----------------------------------------------------------------------

      ! Local variables
      TYPE(iceberg)         , POINTER     :: tmpberg, this
      TYPE(point)           , POINTER     :: pt
      INTEGER                             :: nbergs_to_send_e, nbergs_to_send_w
      INTEGER                             :: nbergs_to_send_n, nbergs_to_send_s
      INTEGER                             :: nbergs_rcvd_from_e, nbergs_rcvd_from_w
      INTEGER                             :: nbergs_rcvd_from_n, nbergs_rcvd_from_s
      INTEGER                             :: i, nbergs_start, nbergs_end
      INTEGER                             :: ine, jne
      INTEGER                             :: pe_N, pe_S, pe_W, pe_E
      REAL(wp), DIMENSION(2)              :: ewbergs, webergs, nsbergs, snbergs
      INTEGER                             :: ml_req1, ml_req2, ml_req3, ml_req4
      INTEGER                             :: ml_req5, ml_req6, ml_req7, ml_req8, ml_err
      INTEGER, DIMENSION(MPI_STATUS_SIZE) :: ml_stat

      ! set up indices of neighbouring processors
      pe_N = -1
      pe_S = -1
      pe_W = -1
      pe_E = -1
      IF( nbondi .EQ.  0 .OR. nbondi .EQ. 1) pe_W = nowe
      IF( nbondi .EQ. -1 .OR. nbondi .EQ. 0) pe_E = noea
      IF( nbondj .EQ.  0 .OR. nbondj .EQ. 1) pe_S = noso
      IF( nbondj .EQ. -1 .OR. nbondj .EQ. 0) pe_N = nono
      !
      ! at northern line of processors with north fold handle bergs differently
      IF( npolj > 0 ) pe_N = -1

      ! if there's only one processor in x direction then don't let mpp try to handle periodicity
      IF( jpni == 1 ) THEN
         pe_E = -1
         pe_W = -1
      ENDIF

      IF( nn_verbose_level >= 2 ) THEN
         WRITE(numicb,*) 'processor west  : ', pe_W
         WRITE(numicb,*) 'processor east  : ', pe_E
         WRITE(numicb,*) 'processor north : ', pe_N
         WRITE(numicb,*) 'processor south : ', pe_S
         WRITE(numicb,*) 'processor nimpp : ', nimpp
         WRITE(numicb,*) 'processor njmpp : ', njmpp
         WRITE(numicb,*) 'processor nbondi: ', nbondi
         WRITE(numicb,*) 'processor nbondj: ', nbondj
         CALL flush( numicb )
      ENDIF

      ! periodicity is handled here when using mpp when there is more than one processor in
      ! the i direction, but it also has to happen when jpni=1 case so this is dealt with
      ! in lbc_send_bergs and called here

      IF( jpni == 1 ) CALL lbc_send_bergs()

      ! Note that xi is adjusted when swapping because of periodic condition

      IF( nn_verbose_level > 0 ) THEN
         ! store the number of icebergs on this processor at start
         nbergs_start = count_bergs()
      ENDIF

      nbergs_to_send_e = 0
      nbergs_to_send_w = 0
      nbergs_to_send_n = 0
      nbergs_to_send_s = 0
      nbergs_rcvd_from_e = 0
      nbergs_rcvd_from_w = 0
      nbergs_rcvd_from_n = 0
      nbergs_rcvd_from_s = 0

      ! Find number of bergs that headed east/west
      IF( ASSOCIATED(first_berg) ) THEN
         this => first_berg
         DO WHILE (ASSOCIATED(this))
            pt => this%current_point
            ine = INT( pt%xi + 0.5 )
            IF( pe_E >= 0 .AND. ine .GT. nimpp+icbei-1 ) THEN
               tmpberg => this
               this => this%next
               nbergs_to_send_e = nbergs_to_send_e + 1
               IF( nn_verbose_level >= 4 ) THEN
                  WRITE(numicb,*) 'bergstep ',ktberg,' packing berg ',tmpberg%number(:),' for transfer to east'
                  CALL flush( numicb )
               ENDIF
               ! deal with periodic case
               tmpberg%current_point%xi = icb_right + MOD(tmpberg%current_point%xi, 1._wp ) - 1._wp
               ! now pack it into buffer and delete from list
               CALL pack_berg_into_buffer( tmpberg, obuffer_e, nbergs_to_send_e)
               CALL delete_iceberg_from_list(first_berg, tmpberg)
            ELSE IF( pe_W >= 0 .AND. ine .LT. nimpp+icbdi-1 ) THEN
               tmpberg => this
               this => this%next
               nbergs_to_send_w = nbergs_to_send_w + 1
               IF( nn_verbose_level >= 4 ) THEN
                  WRITE(numicb,*) 'bergstep ',ktberg,' packing berg ',tmpberg%number(:),' for transfer to west'
                  CALL flush( numicb )
               ENDIF
               ! deal with periodic case
               tmpberg%current_point%xi = icb_left + MOD(tmpberg%current_point%xi, 1._wp )
               ! now pack it into buffer and delete from list
               CALL pack_berg_into_buffer( tmpberg, obuffer_w, nbergs_to_send_w)
               CALL delete_iceberg_from_list(first_berg, tmpberg)
            ELSE
               this => this%next
            ENDIF
         ENDDO
      ENDIF
      if( nn_verbose_level >= 3) then
         write(numicb,*) 'bergstep ',ktberg,' send ew: ', nbergs_to_send_e, nbergs_to_send_w
         call flush(numicb)
      endif

      ! send bergs east and receive bergs from west (ie ones that were sent east)
      ! and vice versa

      ! pattern here is copied from lib_mpp code

      SELECT CASE ( nbondi )
      CASE( -1 )
         webergs(1) = nbergs_to_send_e
         CALL mppsend( 12, webergs(1), 1, pe_E, ml_req1)
         CALL mpprecv( 11, ewbergs(2), 1 )
         IF( l_isend ) CALL mpi_wait( ml_req1, ml_stat, ml_err )
         nbergs_rcvd_from_e = INT( ewbergs(2) )
      CASE(  0 )
         ewbergs(1) = nbergs_to_send_w
         webergs(1) = nbergs_to_send_e
         CALL mppsend( 11, ewbergs(1), 1, pe_W, ml_req2)
         CALL mppsend( 12, webergs(1), 1, pe_E, ml_req3)
         CALL mpprecv( 11, ewbergs(2), 1 )
         CALL mpprecv( 12, webergs(2), 1 )
         IF( l_isend ) CALL mpi_wait( ml_req2, ml_stat, ml_err )
         IF( l_isend ) CALL mpi_wait( ml_req3, ml_stat, ml_err )
         nbergs_rcvd_from_e = INT( ewbergs(2) )
         nbergs_rcvd_from_w = INT( webergs(2) )
      CASE(  1 )
         ewbergs(1) = nbergs_to_send_w
         CALL mppsend( 11, ewbergs(1), 1, pe_W, ml_req4)
         CALL mpprecv( 12, webergs(2), 1 )
         IF( l_isend ) CALL mpi_wait( ml_req4, ml_stat, ml_err )
         nbergs_rcvd_from_w = INT( webergs(2) )
      END SELECT
      if( nn_verbose_level >= 3) then
         write(numicb,*) 'bergstep ',ktberg,' recv ew: ', nbergs_rcvd_from_w, nbergs_rcvd_from_e
         call flush(numicb)
      endif

      SELECT CASE ( nbondi )
      CASE( -1 )
         IF( nbergs_to_send_e > 0 ) CALL mppsend( 14, obuffer_e%data, nbergs_to_send_e*buffer_width, pe_E, ml_req1 )
         IF( nbergs_rcvd_from_e > 0 ) THEN
            CALL increase_ibuffer(ibuffer_e, nbergs_rcvd_from_e)
            CALL mpprecv( 13, ibuffer_e%data, nbergs_rcvd_from_e*buffer_width )
         ENDIF
         IF( nbergs_to_send_e > 0 .AND. l_isend ) CALL mpi_wait( ml_req1, ml_stat, ml_err )
         DO i = 1, nbergs_rcvd_from_e
            IF( nn_verbose_level >= 4 ) THEN
               WRITE(numicb,*) 'bergstep ',ktberg,' unpacking berg ',INT(ibuffer_e%data(16,i)),' from east'
               CALL flush( numicb )
            ENDIF
            CALL unpack_berg_from_buffer(first_berg, ibuffer_e, i)
         ENDDO
      CASE(  0 )
         IF( nbergs_to_send_w > 0 ) CALL mppsend( 13, obuffer_w%data, nbergs_to_send_w*buffer_width, pe_W, ml_req2 )
         IF( nbergs_to_send_e > 0 ) CALL mppsend( 14, obuffer_e%data, nbergs_to_send_e*buffer_width, pe_E, ml_req3 )
         IF( nbergs_rcvd_from_e > 0 ) THEN
            CALL increase_ibuffer(ibuffer_e, nbergs_rcvd_from_e)
            CALL mpprecv( 13, ibuffer_e%data, nbergs_rcvd_from_e*buffer_width )
         ENDIF
         IF( nbergs_rcvd_from_w > 0 ) THEN
            CALL increase_ibuffer(ibuffer_w, nbergs_rcvd_from_w)
            CALL mpprecv( 14, ibuffer_w%data, nbergs_rcvd_from_w*buffer_width )
         ENDIF
         IF( nbergs_to_send_w > 0 .AND. l_isend ) CALL mpi_wait( ml_req2, ml_stat, ml_err )
         IF( nbergs_to_send_e > 0 .AND. l_isend ) CALL mpi_wait( ml_req3, ml_stat, ml_err )
         DO i = 1, nbergs_rcvd_from_e
            IF( nn_verbose_level >= 4 ) THEN
               WRITE(numicb,*) 'bergstep ',ktberg,' unpacking berg ',INT(ibuffer_e%data(16,i)),' from east'
               CALL flush( numicb )
            ENDIF
            CALL unpack_berg_from_buffer(first_berg, ibuffer_e, i)
         ENDDO
         DO i = 1, nbergs_rcvd_from_w
            IF( nn_verbose_level >= 4 ) THEN
               WRITE(numicb,*) 'bergstep ',ktberg,' unpacking berg ',INT(ibuffer_w%data(16,i)),' from west'
               CALL flush( numicb )
            ENDIF
            CALL unpack_berg_from_buffer(first_berg, ibuffer_w, i)
         ENDDO
      CASE(  1 )
         IF( nbergs_to_send_w > 0 ) CALL mppsend( 13, obuffer_w%data, nbergs_to_send_w*buffer_width, pe_W, ml_req4 )
         IF( nbergs_rcvd_from_w > 0 ) THEN
            CALL increase_ibuffer(ibuffer_w, nbergs_rcvd_from_w)
            CALL mpprecv( 14, ibuffer_w%data, nbergs_rcvd_from_w*buffer_width )
         ENDIF
         IF( nbergs_to_send_w > 0 .AND. l_isend ) CALL mpi_wait( ml_req4, ml_stat, ml_err )
         DO i = 1, nbergs_rcvd_from_w
            IF( nn_verbose_level >= 4 ) THEN
               WRITE(numicb,*) 'bergstep ',ktberg,' unpacking berg ',INT(ibuffer_w%data(16,i)),' from west'
               CALL flush( numicb )
            ENDIF
            CALL unpack_berg_from_buffer(first_berg, ibuffer_w, i)
         ENDDO
      END SELECT

      ! Find number of bergs that headed north/south
      ! (note: this block should technically go ahead of the E/W recv block above
      !  to handle arbitrary orientation of PEs. But for simplicity, it is
      !  here to accomodate diagonal transfer of bergs between PEs -AJA)

      IF( ASSOCIATED(first_berg) ) THEN
         this => first_berg
         DO WHILE (ASSOCIATED(this))
            pt => this%current_point
            jne = INT( pt%yj + 0.5 )
            IF( pe_N >= 0 .AND. jne .GT. njmpp+icbej-1 ) THEN
               tmpberg => this
               this => this%next
               nbergs_to_send_n = nbergs_to_send_n + 1
               IF( nn_verbose_level >= 4 ) THEN
                  WRITE(numicb,*) 'bergstep ',ktberg,' packing berg ',tmpberg%number(:),' for transfer to north'
                  CALL flush( numicb )
               ENDIF
               CALL pack_berg_into_buffer( tmpberg, obuffer_n, nbergs_to_send_n)
               CALL delete_iceberg_from_list(first_berg, tmpberg)
            ELSE IF( pe_S >= 0 .AND. jne .LT. njmpp+icbdj-1 ) THEN
               tmpberg => this
               this => this%next
               nbergs_to_send_s = nbergs_to_send_s + 1
               IF( nn_verbose_level >= 4 ) THEN
                  WRITE(numicb,*) 'bergstep ',ktberg,' packing berg ',tmpberg%number(:),' for transfer to south'
                  CALL flush( numicb )
               ENDIF
               CALL pack_berg_into_buffer( tmpberg, obuffer_s, nbergs_to_send_s)
               CALL delete_iceberg_from_list(first_berg, tmpberg)
            ELSE
               this => this%next
            ENDIF
         ENDDO
      ENDIF
      if( nn_verbose_level >= 3) then
         write(numicb,*) 'bergstep ',ktberg,' send ns: ', nbergs_to_send_n, nbergs_to_send_s
         call flush(numicb)
      endif

      ! send bergs north
      ! and receive bergs from south (ie ones sent north)

      SELECT CASE ( nbondj )
      CASE( -1 )
         snbergs(1) = nbergs_to_send_n
         CALL mppsend( 16, snbergs(1), 1, pe_N, ml_req1)
         CALL mpprecv( 15, nsbergs(2), 1 )
         IF( l_isend ) CALL mpi_wait( ml_req1, ml_stat, ml_err )
         nbergs_rcvd_from_n = INT( nsbergs(2) )
      CASE(  0 )
         nsbergs(1) = nbergs_to_send_s
         snbergs(1) = nbergs_to_send_n
         CALL mppsend( 15, nsbergs(1), 1, pe_S, ml_req2)
         CALL mppsend( 16, snbergs(1), 1, pe_N, ml_req3)
         CALL mpprecv( 15, nsbergs(2), 1 )
         CALL mpprecv( 16, snbergs(2), 1 )
         IF( l_isend ) CALL mpi_wait( ml_req2, ml_stat, ml_err )
         IF( l_isend ) CALL mpi_wait( ml_req3, ml_stat, ml_err )
         nbergs_rcvd_from_n = INT( nsbergs(2) )
         nbergs_rcvd_from_s = INT( snbergs(2) )
      CASE(  1 )
         nsbergs(1) = nbergs_to_send_s
         CALL mppsend( 15, nsbergs(1), 1, pe_S, ml_req4)
         CALL mpprecv( 16, snbergs(2), 1 )
         IF( l_isend ) CALL mpi_wait( ml_req4, ml_stat, ml_err )
         nbergs_rcvd_from_s = INT( snbergs(2) )
      END SELECT
      if( nn_verbose_level >= 3) then
         write(numicb,*) 'bergstep ',ktberg,' recv ns: ', nbergs_rcvd_from_s, nbergs_rcvd_from_n
         call flush(numicb)
      endif

      SELECT CASE ( nbondj )
      CASE( -1 )
         IF( nbergs_to_send_n > 0 ) CALL mppsend( 18, obuffer_n%data, nbergs_to_send_n*buffer_width, pe_N, ml_req1 )
         IF( nbergs_rcvd_from_n > 0 ) THEN
            CALL increase_ibuffer(ibuffer_n, nbergs_rcvd_from_n)
            CALL mpprecv( 17, ibuffer_n%data, nbergs_rcvd_from_n*buffer_width )
         ENDIF
         IF( nbergs_to_send_n > 0 .AND. l_isend ) CALL mpi_wait( ml_req1, ml_stat, ml_err )
         DO i = 1, nbergs_rcvd_from_n
            IF( nn_verbose_level >= 4 ) THEN
               WRITE(numicb,*) 'bergstep ',ktberg,' unpacking berg ',INT(ibuffer_n%data(16,i)),' from north'
               CALL flush( numicb )
            ENDIF
            CALL unpack_berg_from_buffer(first_berg, ibuffer_n, i)
         ENDDO
      CASE(  0 )
         IF( nbergs_to_send_s > 0 ) CALL mppsend( 17, obuffer_s%data, nbergs_to_send_s*buffer_width, pe_S, ml_req2 )
         IF( nbergs_to_send_n > 0 ) CALL mppsend( 18, obuffer_n%data, nbergs_to_send_n*buffer_width, pe_N, ml_req3 )
         IF( nbergs_rcvd_from_n > 0 ) THEN
            CALL increase_ibuffer(ibuffer_n, nbergs_rcvd_from_n)
            CALL mpprecv( 17, ibuffer_n%data, nbergs_rcvd_from_n*buffer_width )
         ENDIF
         IF( nbergs_rcvd_from_s > 0 ) THEN
            CALL increase_ibuffer(ibuffer_s, nbergs_rcvd_from_s)
            CALL mpprecv( 18, ibuffer_s%data, nbergs_rcvd_from_s*buffer_width )
         ENDIF
         IF( nbergs_to_send_s > 0 .AND. l_isend ) CALL mpi_wait( ml_req2, ml_stat, ml_err )
         IF( nbergs_to_send_n > 0 .AND. l_isend ) CALL mpi_wait( ml_req3, ml_stat, ml_err )
         DO i = 1, nbergs_rcvd_from_n
            IF( nn_verbose_level >= 4 ) THEN
               WRITE(numicb,*) 'bergstep ',ktberg,' unpacking berg ',INT(ibuffer_n%data(16,i)),' from north'
               CALL flush( numicb )
            ENDIF
            CALL unpack_berg_from_buffer(first_berg, ibuffer_n, i)
         ENDDO
         DO i = 1, nbergs_rcvd_from_s
            IF( nn_verbose_level >= 4 ) THEN
               WRITE(numicb,*) 'bergstep ',ktberg,' unpacking berg ',INT(ibuffer_s%data(16,i)),' from south'
               CALL flush( numicb )
            ENDIF
            CALL unpack_berg_from_buffer(first_berg, ibuffer_s, i)
         ENDDO
      CASE(  1 )
         IF( nbergs_to_send_s > 0 ) CALL mppsend( 17, obuffer_s%data, nbergs_to_send_s*buffer_width, pe_S, ml_req4 )
         IF( nbergs_rcvd_from_s > 0 ) THEN
            CALL increase_ibuffer(ibuffer_s, nbergs_rcvd_from_s)
            CALL mpprecv( 18, ibuffer_s%data, nbergs_rcvd_from_s*buffer_width )
         ENDIF
         IF( nbergs_to_send_s > 0 .AND. l_isend ) CALL mpi_wait( ml_req4, ml_stat, ml_err )
         DO i = 1, nbergs_rcvd_from_s
            IF( nn_verbose_level >= 4 ) THEN
               WRITE(numicb,*) 'bergstep ',ktberg,' unpacking berg ',INT(ibuffer_s%data(16,i)),' from south'
               CALL flush( numicb )
            ENDIF
            CALL unpack_berg_from_buffer(first_berg, ibuffer_s, i)
         ENDDO
      END SELECT

      IF( nn_verbose_level > 0 ) THEN
         ! compare the number of icebergs on this processor from the start to the end
         nbergs_end = count_bergs()
         i = ( nbergs_rcvd_from_n + nbergs_rcvd_from_s + nbergs_rcvd_from_e + nbergs_rcvd_from_w ) - &
             ( nbergs_to_send_n + nbergs_to_send_s + nbergs_to_send_e + nbergs_to_send_w )
         IF( nbergs_end-(nbergs_start+i) .NE. 0 ) THEN
            WRITE( numicb,*   ) 'send_bergs_to_other_pes: net change in number of icebergs'
            WRITE( numicb,1000) 'send_bergs_to_other_pes: nbergs_end=', &
                                nbergs_end,' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: nbergs_start=', &
                                nbergs_start,' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: delta=', &
                                i,' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: error=', &
                                nbergs_end-(nbergs_start+i),' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: nbergs_to_send_n=', &
                                nbergs_to_send_n,' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: nbergs_to_send_s=', &
                                nbergs_to_send_s,' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: nbergs_to_send_e=', &
                                nbergs_to_send_e,' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: nbergs_to_send_w=', &
                                nbergs_to_send_w,' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: nbergs_rcvd_from_n=', &
                                nbergs_rcvd_from_n,' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: nbergs_rcvd_from_s=', &
                                nbergs_rcvd_from_s,' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: nbergs_rcvd_from_e=', &
                                nbergs_rcvd_from_e,' on PE',narea
            WRITE( numicb,1000) 'send_bergs_to_other_pes: nbergs_rcvd_from_w=', &
                                nbergs_rcvd_from_w,' on PE',narea
  1000      FORMAT(a,i5,a,i4)
            CALL ctl_stop('send_bergs_to_other_pes: lost or gained an iceberg or two')
         ENDIF
      ENDIF

      ! deal with north fold if we necessary when there is more than one top row processor
      ! note that for jpni=1 north fold has been dealt with above in call to lbc_send_bergs
      IF( npolj /= 0 .AND. jpni > 1 ) CALL mpp_nfld_bergs( )

      IF( nn_verbose_level > 0 ) THEN
         i = 0
         this => first_berg
         DO WHILE (ASSOCIATED(this))
            pt => this%current_point
            ine = INT( pt%xi + 0.5 )
            jne = INT( pt%yj + 0.5 )
!           CALL check_position(grd, this, 'exchange (bot)')
            IF( ine .LT. nimpp+icbdi-1 .OR. &
                ine .GT. nimpp+icbei-1 .OR. &
                jne .LT. njmpp+icbdj-1 .OR. &
                jne .GT. njmpp+icbej-1) THEN
               i = i + 1
               WRITE(numicb,*) 'berg lost in halo: ', this%number(:),ine,jne
               WRITE(numicb,*) '                   ', nimpp, njmpp
               WRITE(numicb,*) '                   ', icbdi, icbei, icbdj, icbej
               CALL flush( numicb )
            ENDIF
            this => this%next
         ENDDO ! WHILE
         CALL mpp_sum(i)
         IF( i .GT. 0 ) THEN
            WRITE( numicb,'(a,i4)') 'send_bergs_to_other_pes: # of bergs outside computational domain = ',i
            CALL ctl_stop('send_bergs_to_other_pes:  there are bergs still in halos!')
         ENDIF ! root_pe
      ENDIF ! debug

      CALL mppsync()

   END SUBROUTINE mpp_send_bergs

   !!-------------------------------------------------------------------------

   SUBROUTINE mpp_nfld_bergs()

      ! Local variables
      TYPE(iceberg)         , POINTER     :: tmpberg, this
      TYPE(point)           , POINTER     :: pt
      INTEGER                             :: nbergs_to_send
      INTEGER                             :: nbergs_to_rcv
      INTEGER                             :: jiglo, jjglo, jk, jn
      INTEGER                             :: jfldproc, kproc, kpts
      INTEGER                             :: ine, jne
      REAL(wp), DIMENSION(2)              :: sbergs, nbergs
      INTEGER                             :: ml_req1, ml_req2, ml_err
      INTEGER, DIMENSION(MPI_STATUS_SIZE) :: ml_stat

      ! set up indices of neighbouring processors

      ! icbfldproc is a list of unique processor numbers that this processor
      ! exchanges with (including itself), so we loop over this array; since
      ! its of fixed size, the first -1 marks end of list of processors
      !
      DO jn = 1, jpni
         IF( icbfldproc(jn) == -1 ) EXIT
         jfldproc = icbfldproc(jn)
         nbergs_to_send = 0

         ! Find number of bergs that need to be exchanged
         ! Pick out exchanges with processor jfldproc
         ! if jfldproc is this processor then don't send
         !
         IF( ASSOCIATED(first_berg) ) THEN
            this => first_berg
            DO WHILE (ASSOCIATED(this))
               pt => this%current_point
               ine = INT( pt%xi + 0.5 )
               jne = INT( pt%yj + 0.5 )
               kpts  = icbfldpts (ine-nimpp+1)
               kproc = icbflddest(ine-nimpp+1)
               IF( jne .GT. njmpp+icbej-1 ) THEN
                  IF( kproc == jfldproc ) THEN
                     !
                     ! moving across the cut line means both position and
                     ! velocity must change
                     jjglo = INT( kpts/icbpack )
                     jiglo = kpts - icbpack*jjglo
                     pt%xi = jiglo - ( pt%xi - REAL(ine,wp) )
                     pt%yj = jjglo - ( pt%yj - REAL(jne,wp) )
                     pt%uvel = -1._wp * pt%uvel
                     pt%vvel = -1._wp * pt%vvel
                     !
                     ! now remove berg from list and pack it into a buffer
                     IF( kproc /= narea ) THEN
                        tmpberg => this
                        nbergs_to_send = nbergs_to_send + 1
                        IF( nn_verbose_level >= 4 ) THEN
                           WRITE(numicb,*) 'bergstep ',ktberg,' packing berg ',tmpberg%number(:),' for north fold'
                           CALL flush( numicb )
                        ENDIF
                        CALL pack_berg_into_buffer( tmpberg, obuffer_f, nbergs_to_send)
                        CALL delete_iceberg_from_list(first_berg, tmpberg)
                     ENDIF
                     !
                  ENDIF
               ENDIF
               this => this%next
            ENDDO
         ENDIF
         if( nn_verbose_level >= 3) then
            write(numicb,*) 'bergstep ',ktberg,' send nfld: ', nbergs_to_send
            call flush(numicb)
         endif
         !
         ! if we're in this processor, then we've done everything we need to
         ! so go on to next element of loop
         IF( jfldproc == narea ) CYCLE

         sbergs(1) = nbergs_to_send
         CALL mppsend( 21, sbergs(1), 1, jfldproc-1, ml_req1)
         CALL mpprecv( 21, nbergs(2), 1 )
         IF( l_isend ) CALL mpi_wait( ml_req1, ml_stat, ml_err )
         nbergs_to_rcv = INT( nbergs(2) )

         ! send bergs

         IF( nbergs_to_send > 0 )  &
             CALL mppsend( 12, obuffer_f%data, nbergs_to_send*buffer_width, jfldproc-1, ml_req2 )
         IF( nbergs_to_rcv  > 0 ) THEN
            CALL increase_ibuffer(ibuffer_f, nbergs_to_rcv)
            CALL mpprecv( 12, ibuffer_f%data, nbergs_to_rcv*buffer_width )
         ENDIF
         IF( nbergs_to_send > 0 .AND. l_isend ) CALL mpi_wait( ml_req2, ml_stat, ml_err )
         DO jk = 1, nbergs_to_rcv
            IF( nn_verbose_level >= 4 ) THEN
               WRITE(numicb,*) 'bergstep ',ktberg,' unpacking berg ',INT(ibuffer_f%data(16,jk)),' from north fold'
               CALL flush( numicb )
            ENDIF
            CALL unpack_berg_from_buffer(first_berg, ibuffer_f, jk )
         ENDDO

      ENDDO

   END SUBROUTINE mpp_nfld_bergs

   !!-------------------------------------------------------------------------

   SUBROUTINE dealloc_buffers()

      CALL dealloc_buffer( obuffer_n )
      CALL dealloc_buffer( obuffer_s )
      CALL dealloc_buffer( obuffer_e )
      CALL dealloc_buffer( obuffer_w )
      CALL dealloc_buffer( ibuffer_n )
      CALL dealloc_buffer( ibuffer_s )
      CALL dealloc_buffer( ibuffer_e )
      CALL dealloc_buffer( ibuffer_w )

   END SUBROUTINE dealloc_buffers

   !!-------------------------------------------------------------------------

   SUBROUTINE dealloc_buffer(buff)
      ! Arguments
      TYPE(buffer), POINTER :: buff

      IF( ASSOCIATED(buff) ) THEN
         IF( ASSOCIATED(buff%data)) DEALLOCATE(buff%data)
         DEALLOCATE(buff)
      ENDIF
   END SUBROUTINE dealloc_buffer

   !!-------------------------------------------------------------------------

   SUBROUTINE pack_berg_into_buffer(berg, buff, n)
      ! Arguments
      TYPE(iceberg),            POINTER :: berg
      TYPE(buffer) ,            POINTER :: buff
      INTEGER      , INTENT(in)         :: n
      ! Local
      INTEGER                           :: k

      IF( .NOT. ASSOCIATED(buff) ) CALL increase_buffer( buff, delta_buf )
      IF( n .GT. buff%size ) CALL increase_buffer( buff, delta_buf )

      !! pack points into buffer

      buff%data( 1,n) = berg%current_point%lon
      buff%data( 2,n) = berg%current_point%lat
      buff%data( 3,n) = berg%current_point%uvel
      buff%data( 4,n) = berg%current_point%vvel
      buff%data( 5,n) = berg%current_point%xi
      buff%data( 6,n) = berg%current_point%yj
      buff%data( 7,n) = float(berg%current_point%year)
      buff%data( 8,n) = berg%current_point%day
      buff%data( 9,n) = berg%current_point%mass
      buff%data(10,n) = berg%current_point%thickness
      buff%data(11,n) = berg%current_point%width
      buff%data(12,n) = berg%current_point%length
      buff%data(13,n) = berg%current_point%mass_of_bits
      buff%data(14,n) = berg%current_point%heat_density

      buff%data(15,n) = berg%mass_scaling
      DO k=1,nkounts
         buff%data(15+k,n) = REAL( berg%number(k), wp )
      END DO

   END SUBROUTINE pack_berg_into_buffer

   !!-------------------------------------------------------------------------

   SUBROUTINE unpack_berg_from_buffer(first, buff, n)
      ! Arguments
      TYPE(iceberg),             POINTER :: first
      TYPE(buffer) ,             POINTER :: buff
      INTEGER      , INTENT(in)          :: n
      ! Local variables
      LOGICAL                            :: lres
      TYPE(iceberg)                      :: currentberg
      TYPE(point)                        :: pt
      INTEGER                           :: k

      pt%lon            =      buff%data( 1,n)
      pt%lat            =      buff%data( 2,n)
      pt%uvel           =      buff%data( 3,n)
      pt%vvel           =      buff%data( 4,n)
      pt%xi             =      buff%data( 5,n)
      pt%yj             =      buff%data( 6,n)
      pt%year           = INT( buff%data( 7,n) )
      pt%day            =      buff%data( 8,n)
      pt%mass           =      buff%data( 9,n)
      pt%thickness      =      buff%data(10,n)
      pt%width          =      buff%data(11,n)
      pt%length         =      buff%data(12,n)
      pt%mass_of_bits   =      buff%data(13,n)
      pt%heat_density   =      buff%data(14,n)

      currentberg%mass_scaling =      buff%data(15,n)
      DO k=1,nkounts
         currentberg%number(k) = INT( buff%data(15+k,n) )
      END DO

      CALL add_new_berg_to_list(currentberg, pt )

   END SUBROUTINE unpack_berg_from_buffer

   !!-------------------------------------------------------------------------

   SUBROUTINE increase_buffer(old,delta)
      ! Arguments
      TYPE(buffer),             POINTER :: old
      INTEGER     , INTENT(in)          :: delta
      ! Local variables
      TYPE(buffer),             POINTER :: new
      INTEGER                           :: new_size

      IF( .NOT. ASSOCIATED(old) ) THEN
         new_size = delta
      ELSE
         new_size = old%size + delta
      ENDIF
      ALLOCATE( new )
      ALLOCATE( new%data( buffer_width, new_size) )
      new%size = new_size
      IF( ASSOCIATED(old) ) THEN
         new%data(:,1:old%size) = old%data(:,1:old%size)
         DEALLOCATE(old%data)
         DEALLOCATE(old)
      ENDIF
      old => new

   END SUBROUTINE increase_buffer

   !!-------------------------------------------------------------------------

   SUBROUTINE increase_ibuffer(old,delta)
      ! Arguments
      TYPE(buffer),            POINTER :: old
      INTEGER     , INTENT(in)         :: delta
      ! Local variables
      TYPE(buffer),            POINTER :: new
      INTEGER                          :: new_size, old_size

      IF( .NOT. ASSOCIATED(old) ) THEN
         new_size = delta + delta_buf
         old_size = 0
      ELSE
         old_size = old%size
         IF( delta .LT. old%size ) THEN
            new_size = old%size + delta
         ELSE
            new_size = delta + delta_buf
         ENDIF
      ENDIF

      IF( old_size .NE. new_size ) THEN
         ALLOCATE( new )
         ALLOCATE( new%data( buffer_width, new_size) )
         new%size = new_size
         IF( ASSOCIATED(old) ) THEN
            new%data(:,1:old%size) = old%data(:,1:old%size)
            DEALLOCATE(old%data)
            DEALLOCATE(old)
         ENDIF
         old => new
        !WRITE( numicb,*) 'increase_ibuffer',narea,' increased to',new_size
      ENDIF

   END SUBROUTINE increase_ibuffer

   !!-------------------------------------------------------------------------

#else

   SUBROUTINE mpp_send_bergs()
      WRITE(numout,*) 'mpp_send_bergs: You should not have seen this message!!'
   END SUBROUTINE mpp_send_bergs

   SUBROUTINE dealloc_buffers()
      WRITE(numout,*) 'dealloc_buffers: You should not have seen this message!!'
   END SUBROUTINE dealloc_buffers

#endif

END MODULE icblbc
