MODULE icbrst

   !!======================================================================
   !!                       ***  MODULE  icbrst  ***
   !! Ocean physics:  read and write iceberg restart files
   !!======================================================================
   !! History : 3.3.1 !  2010-01  (Martin&Adcroft) Original code
   !!            -    !  2011-03  (Madec)          Part conversion to NEMO form
   !!            -    !                            Removal of mapping from another grid
   !!            -    !  2011-04  (Alderson)       Split into separate modules
   !!            -    !  2011-04  (Alderson)       Restore restart routine
   !!            -    !                            Currently needs a fixed processor
   !!            -    !                            layout between restarts
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   icebergs_read_restart    : initialise
   !!   icebergs_write_restart   : generate test icebergs
   !!----------------------------------------------------------------------
   USE par_oce        ! NEMO parameters
   USE dom_oce        ! NEMO domain
   USE in_out_manager ! NEMO IO routines
   USE lib_mpp        ! NEMO MPI library, lk_mpp in particular
   USE netcdf         ! netcdf routines for IO
   USE icb_oce        ! define iceberg arrays
   USE icbutl         ! iceberg utility routines

   IMPLICIT NONE
   PRIVATE

   REAL(wp), DIMENSION(:,:,:), PRIVATE, POINTER :: griddata => NULL()    ! need 2d array to read in with
   INTEGER,  DIMENSION(3),     PRIVATE          :: strt3, lngth3
   INTEGER,                    PRIVATE          :: lonid, latid, xid, yid, uvelid, vvelid
   INTEGER,                    PRIVATE          :: massid, thicknessid, widthid, lengthid
   INTEGER,                    PRIVATE          :: yearid, dayid
   INTEGER,                    PRIVATE          :: scaling_id, mass_of_bits_id, heat_density_id, numberid
   INTEGER,                    PRIVATE          :: siceid, sheatid, calvid, calvhid, kountid

   PUBLIC   icebergs_read_restart       ! routine called in xxx.F90 module
   PUBLIC   icebergs_write_restart      ! routine called in xxx.F90 module

CONTAINS

   SUBROUTINE icebergs_read_restart()
      ! sga - read a restart file
      !       for this version, we just read back in the restart for this processor
      !       so we cannot change the processor layout currently with iceberg code

      ! Local variables
      INTEGER                               :: ndim, nvar, natt
      INTEGER                               :: k, iret, ncid, unlim_dim, x_dim, y_dim, c_dim, nbergs_in_file
      INTEGER                               :: mclass
      INTEGER, DIMENSION(1)                 :: strt, lngth, idata
      INTEGER, DIMENSION(2)                 :: strt2, lngth2
      INTEGER, DIMENSION(nkounts)           :: idata2
      REAL(wp), DIMENSION(1)                :: data                                         ! need 1d array to read in with
                                                                                            ! start and count arrays
      LOGICAL                               :: lres, found_restart, multiPErestart=.FALSE.
      CHARACTER(len=80)                     :: filename
      CHARACTER(len=NF90_MAX_NAME)          :: dname
      TYPE(iceberg)                         :: localberg ! NOT a pointer but an actual local variable
      TYPE(point)                           :: localpt   ! NOT a pointer but an actual local variable

      IF( .NOT. ASSOCIATED(griddata) ) ALLOCATE( griddata(jpi,jpj,1) )

      ! Find a restart file
      multiPErestart=.FALSE.
      DO
         filename = ' '
         filename = 'restart_icebergs.nc'
         INQUIRE(file=TRIM(filename),exist=found_restart)
         IF (found_restart) EXIT
         filename = ' '
         WRITE(filename,'("restart_icebergs_",I4.4,".nc")') narea-1
         INQUIRE(file=TRIM(filename),exist=found_restart)
         IF (found_restart) THEN
            multiPErestart=.TRUE.
            EXIT
         ENDIF
         IF (nn_verbose_level >= 0 .AND. lwp) &
            WRITE(numout,'(a)') 'read_restart_bergs: no restart file found'
         multiPErestart=.TRUE. ! force sanity checking in a multi-PE mode if no file found on this PE
         EXIT
      ENDDO

      IF ( .NOT. found_restart) THEN ! only do the following if a file was found
         CALL ctl_stop('icebergs: no restart file found')
      ENDIF

      IF (nn_verbose_level >= 0 .AND. lwp)  &
         WRITE(numout,'(2a)') 'icebergs, read_restart_bergs: found restart file = ',TRIM(filename)

      iret = NF90_OPEN(TRIM(filename), NF90_NOWRITE, ncid)
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, read_restart_bergs: nf_open failed')

      iret = nf90_inquire(ncid, ndim, nvar, natt, unlim_dim)
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, read_restart_bergs: nf_inquire failed')

      IF( unlim_dim .NE. -1) THEN

         iret = nf90_inquire_dimension(ncid, unlim_dim, dname, nbergs_in_file)
         IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, read_restart_bergs: nf_inq_dimlen failed')

         iret = NF90_INQ_VARID(ncid, 'number', numberid)
         iret = NF90_INQ_VARID(ncid, 'mass_scaling', scaling_id)
         iret = NF90_INQ_VARID(ncid, 'xi', xid)
         iret = NF90_INQ_VARID(ncid, 'yj', yid)
         iret = NF90_INQ_VARID(ncid, 'lon', lonid)
         iret = NF90_INQ_VARID(ncid, 'lat', latid)
         iret = NF90_INQ_VARID(ncid, 'uvel', uvelid)
         iret = NF90_INQ_VARID(ncid, 'vvel', vvelid)
         iret = NF90_INQ_VARID(ncid, 'mass', massid)
         iret = NF90_INQ_VARID(ncid, 'thickness', thicknessid)
         iret = NF90_INQ_VARID(ncid, 'width', widthid)
         iret = NF90_INQ_VARID(ncid, 'length', lengthid)
         iret = NF90_INQ_VARID(ncid, 'year', yearid)
         iret = NF90_INQ_VARID(ncid, 'day', dayid)
         iret = NF90_INQ_VARID(ncid, 'mass_of_bits', mass_of_bits_id)
         iret = NF90_INQ_VARID(ncid, 'heat_density', heat_density_id)

         lngth(1) = 1
         strt2(1) = 1
         lngth2(1) = nkounts
         lngth2(2) = 1
         DO k=1, nbergs_in_file

            strt(1) = k
            strt2(2) = k

            iret = NF90_GET_VAR(ncid, numberid, idata2, strt2, lngth2 )
            localberg%number(:) = idata2(:)

            iret = NF90_GET_VAR(ncid, scaling_id, data, strt, lngth )
            localberg%mass_scaling = data(1)

            iret = NF90_GET_VAR(ncid, lonid, data, strt, lngth)
            localpt%lon = data(1)
            iret = NF90_GET_VAR(ncid, latid, data, strt, lngth)
            localpt%lat = data(1)
            IF (nn_verbose_level >= 2 .AND. lwp) THEN
               WRITE(numout,'(a,i5,a,2f10.4,a,i5)') 'icebergs, read_restart_bergs: berg ',k,' is at ', &
                                              localpt%lon,localpt%lat,' on PE ',narea-1
            ENDIF
            iret = NF90_GET_VAR(ncid, xid, data, strt, lngth)
            localpt%xi = data(1)
            iret = NF90_GET_VAR(ncid, yid, data, strt, lngth)
            localpt%yj = data(1)
            iret = NF90_GET_VAR(ncid, uvelid, data, strt, lngth )
            localpt%uvel = data(1)
            iret = NF90_GET_VAR(ncid, vvelid, data, strt, lngth )
            localpt%vvel = data(1)
            iret = NF90_GET_VAR(ncid, massid, data, strt, lngth )
            localpt%mass = data(1)
            iret = NF90_GET_VAR(ncid, thicknessid, data, strt, lngth )
            localpt%thickness = data(1)
            iret = NF90_GET_VAR(ncid, widthid, data, strt, lngth )
            localpt%width = data(1)
            iret = NF90_GET_VAR(ncid, lengthid, data, strt, lngth )
            localpt%length = data(1)
            iret = NF90_GET_VAR(ncid, yearid, idata, strt, lngth )
            localpt%year = idata(1)
            iret = NF90_GET_VAR(ncid, dayid, data, strt, lngth )
            localpt%day = data(1)
            iret = NF90_GET_VAR(ncid, mass_of_bits_id, data, strt, lngth )
            localpt%mass_of_bits = data(1)
            iret = NF90_GET_VAR(ncid, heat_density_id, data, strt, lngth )
            localpt%heat_density = data(1)
            !
            CALL add_new_berg_to_list( localberg, localpt )
         ENDDO

      ENDIF

      iret = NF90_INQ_DIMID( ncid, 'c', c_dim )
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, read_restart: nf_inq_dimid c failed')

      iret = NF90_INQUIRE_DIMENSION( ncid, c_dim, dname, mclass )
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, read_restart: nf_inquire_dimension failed')

      iret = NF90_INQ_VARID(ncid, 'kount'       , kountid)
      iret = NF90_INQ_VARID(ncid, 'calving'     , calvid)
      iret = NF90_INQ_VARID(ncid, 'calving_hflx', calvhid)
      iret = NF90_INQ_VARID(ncid, 'stored_ice'  , siceid)
      iret = NF90_INQ_VARID(ncid, 'stored_heat' , sheatid)

      strt3(1) = 1
      strt3(2) = 1
      lngth3(1) = jpi
      lngth3(2) = jpj
      lngth3(3) = 1

      DO k=1,mclass

         strt3(3) = k

         iret = NF90_GET_VAR( ncid, siceid , griddata, strt3, lngth3 )
         berg_grid%stored_ice(:,:,k) = griddata(:,:,1)

      ENDDO

      iret = NF90_GET_VAR( ncid, calvid , p_calving(:,:) )
      iret = NF90_GET_VAR( ncid, calvhid, p_calving_hflx(:,:) )
      iret = NF90_GET_VAR( ncid, sheatid, berg_grid%stored_heat(:,:) )
      iret = NF90_GET_VAR( ncid, kountid, idata2(:) )
      kount_bergs(:) = idata2(:)

      ! Finish up
      iret = NF90_CLOSE(ncid)
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, read_restart: nf_close failed')

      ! Sanity check
      k = count_bergs()
      IF (nn_verbose_level >= 0)   &
         WRITE(numout,'(2(a,i5))') 'icebergs, read_restart_bergs: # bergs =',k,' on PE',narea-1
      IF( lk_mpp ) THEN
         IF (multiPErestart) CALL mpp_sum(nbergs_in_file) ! In case PE 0 didn't open a file
         CALL mpp_sum(k)
      ENDIF
      IF ( lwp ) THEN
         WRITE(numout,'(a,i5,a,i5,a)') 'icebergs, read_restart_bergs: there were',nbergs_in_file, &
                                       ' bergs in the restart file and', k,' bergs have been read'
      ENDIF

      IF ( lwp .and. nn_verbose_level >= 0)  &
         WRITE(numout,'(a)') 'icebergs, read_restart_bergs: completed'

   END SUBROUTINE icebergs_read_restart

! ##############################################################################

   SUBROUTINE icebergs_write_restart( kt )

      INTEGER, INTENT( in )                 :: kt
      ! Local variables
      INTEGER                               :: iret, ncid, k_dim, n_dim, x_dim, y_dim, c_dim, k
      CHARACTER(len=80)                     :: filename
      TYPE(iceberg), POINTER                :: this
      TYPE(point)  , POINTER                :: pt

      IF( .NOT. ASSOCIATED(griddata) ) ALLOCATE( griddata(jpi,jpj,1) )

      WRITE(filename,'("icebergs_",I8.8,"_restart_",I4.4,".nc")') kt, narea-1
      IF (nn_verbose_level >= 0) WRITE(numout,'(2a)') 'icebergs, write_restart: creating ',TRIM(filename)

      iret = NF90_CREATE(TRIM(filename), NF90_CLOBBER, ncid)
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_create failed')

      ! Dimensions
      iret = NF90_DEF_DIM(ncid, 'x', jpi, x_dim)
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_def_dim x failed')

      iret = NF90_DEF_DIM(ncid, 'y', jpj, y_dim)
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_def_dim y failed')

      iret = NF90_DEF_DIM(ncid, 'c', nclasses, c_dim)
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_def_dim c failed')

      iret = NF90_DEF_DIM(ncid, 'k', nkounts, k_dim)
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_def_dim k failed')

      IF (associated(first_berg)) then
         iret = NF90_DEF_DIM(ncid, 'n', NF90_UNLIMITED, n_dim)
         IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_def_dim n failed')
      ENDIF

      ! Variables
      iret = NF90_DEF_VAR(ncid, 'kount'       , NF90_INT   , (/ k_dim /), kountid)
      iret = NF90_DEF_VAR(ncid, 'calving'     , NF90_DOUBLE, (/ x_dim, y_dim /), calvid)
      iret = NF90_DEF_VAR(ncid, 'calving_hflx', NF90_DOUBLE, (/ x_dim, y_dim /), calvhid)
      iret = NF90_DEF_VAR(ncid, 'stored_ice'  , NF90_DOUBLE, (/ x_dim, y_dim, c_dim /), siceid)
      iret = NF90_DEF_VAR(ncid, 'stored_heat' , NF90_DOUBLE, (/ x_dim, y_dim /), sheatid)

      ! Attributes
      iret = NF90_PUT_ATT(ncid, calvid , 'long_name', 'iceberg calving')
      iret = NF90_PUT_ATT(ncid, calvid , 'units', 'some')
      iret = NF90_PUT_ATT(ncid, calvhid, 'long_name', 'heat flux associated with iceberg calving')
      iret = NF90_PUT_ATT(ncid, calvhid, 'units', 'some')
      iret = NF90_PUT_ATT(ncid, siceid , 'long_name', 'stored ice used to calve icebergs')
      iret = NF90_PUT_ATT(ncid, siceid , 'units', 'kg/s')
      iret = NF90_PUT_ATT(ncid, sheatid, 'long_name', 'heat in stored ice used to calve icebergs')
      iret = NF90_PUT_ATT(ncid, sheatid, 'units', 'J/kg/s')

      IF ( ASSOCIATED(first_berg) ) THEN

         ! Only add berg variables for this PE if we have anything to say

         ! Variables
         iret = NF90_DEF_VAR(ncid, 'lon', NF90_DOUBLE, n_dim, lonid)
         iret = NF90_DEF_VAR(ncid, 'lat', NF90_DOUBLE, n_dim, latid)
         iret = NF90_DEF_VAR(ncid, 'xi', NF90_DOUBLE, n_dim, xid)
         iret = NF90_DEF_VAR(ncid, 'yj', NF90_DOUBLE, n_dim, yid)
         iret = NF90_DEF_VAR(ncid, 'uvel', NF90_DOUBLE, n_dim, uvelid)
         iret = NF90_DEF_VAR(ncid, 'vvel', NF90_DOUBLE, n_dim, vvelid)
         iret = NF90_DEF_VAR(ncid, 'mass', NF90_DOUBLE, n_dim, massid)
         iret = NF90_DEF_VAR(ncid, 'thickness', NF90_DOUBLE, n_dim, thicknessid)
         iret = NF90_DEF_VAR(ncid, 'width', NF90_DOUBLE, n_dim, widthid)
         iret = NF90_DEF_VAR(ncid, 'length', NF90_DOUBLE, n_dim, lengthid)
         iret = NF90_DEF_VAR(ncid, 'number', NF90_INT, (/k_dim,n_dim/), numberid)
         iret = NF90_DEF_VAR(ncid, 'year', NF90_INT, n_dim, yearid)
         iret = NF90_DEF_VAR(ncid, 'day', NF90_DOUBLE, n_dim, dayid)
         iret = NF90_DEF_VAR(ncid, 'mass_scaling', NF90_DOUBLE, n_dim, scaling_id)
         iret = NF90_DEF_VAR(ncid, 'mass_of_bits', NF90_DOUBLE, n_dim, mass_of_bits_id)
         iret = NF90_DEF_VAR(ncid, 'heat_density', NF90_DOUBLE, n_dim, heat_density_id)

         ! Attributes
         iret = NF90_PUT_ATT(ncid, lonid, 'long_name', 'longitude')
         iret = NF90_PUT_ATT(ncid, lonid, 'units', 'degrees_E')
         iret = NF90_PUT_ATT(ncid, latid, 'long_name', 'latitude')
         iret = NF90_PUT_ATT(ncid, latid, 'units', 'degrees_N')
         iret = NF90_PUT_ATT(ncid, xid, 'long_name', 'x grid box position')
         iret = NF90_PUT_ATT(ncid, xid, 'units', 'fractional')
         iret = NF90_PUT_ATT(ncid, yid, 'long_name', 'y grid box position')
         iret = NF90_PUT_ATT(ncid, yid, 'units', 'fractional')
         iret = NF90_PUT_ATT(ncid, uvelid, 'long_name', 'zonal velocity')
         iret = NF90_PUT_ATT(ncid, uvelid, 'units', 'm/s')
         iret = NF90_PUT_ATT(ncid, vvelid, 'long_name', 'meridional velocity')
         iret = NF90_PUT_ATT(ncid, vvelid, 'units', 'm/s')
         iret = NF90_PUT_ATT(ncid, massid, 'long_name', 'mass')
         iret = NF90_PUT_ATT(ncid, massid, 'units', 'kg')
         iret = NF90_PUT_ATT(ncid, thicknessid, 'long_name', 'thickness')
         iret = NF90_PUT_ATT(ncid, thicknessid, 'units', 'm')
         iret = NF90_PUT_ATT(ncid, widthid, 'long_name', 'width')
         iret = NF90_PUT_ATT(ncid, widthid, 'units', 'm')
         iret = NF90_PUT_ATT(ncid, lengthid, 'long_name', 'length')
         iret = NF90_PUT_ATT(ncid, lengthid, 'units', 'm')
         iret = NF90_PUT_ATT(ncid, numberid, 'long_name', 'iceberg number on this processor')
         iret = NF90_PUT_ATT(ncid, numberid, 'units', 'count')
         iret = NF90_PUT_ATT(ncid, yearid, 'long_name', 'calendar year of calving event')
         iret = NF90_PUT_ATT(ncid, yearid, 'units', 'years')
         iret = NF90_PUT_ATT(ncid, dayid, 'long_name', 'year day of calving event')
         iret = NF90_PUT_ATT(ncid, dayid, 'units', 'days')
         iret = NF90_PUT_ATT(ncid, scaling_id, 'long_name', 'scaling factor for mass of calving berg')
         iret = NF90_PUT_ATT(ncid, scaling_id, 'units', 'none')
         iret = NF90_PUT_ATT(ncid, mass_of_bits_id, 'long_name', 'mass of bergy bits')
         iret = NF90_PUT_ATT(ncid, mass_of_bits_id, 'units', 'kg')
         iret = NF90_PUT_ATT(ncid, heat_density_id, 'long_name', 'heat density')
         iret = NF90_PUT_ATT(ncid, heat_density_id, 'units', 'J/kg')

      ENDIF ! associated(first_berg)

      ! End define mode
      iret = NF90_ENDDEF(ncid)

      ! --------------------------------
      ! now write some data

      strt3(1) = 1
      strt3(2) = 1
      lngth3(1) = jpi
      lngth3(2) = jpj
      lngth3(3) = 1

      DO k=1,nclasses
         griddata(:,:,1) = berg_grid%stored_ice(:,:,k)
         strt3(3) = k
         iret = NF90_PUT_VAR( ncid, siceid, griddata, strt3, lngth3 )
         IF (iret .ne. NF90_NOERR) THEN
            IF( lwp ) WRITE(numout,*) TRIM(NF90_STRERROR( iret ))
            CALL ctl_stop('icebergs, write_restart: nf_put_var stored_ice failed')
         ENDIF
      ENDDO
      IF( lwp ) WRITE(numout,*) 'file: ',TRIM(filename),' var: stored_ice  written'

      iret = NF90_PUT_VAR( ncid, kountid, kount_bergs(:) )
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_put_var kount failed')

      iret = NF90_PUT_VAR( ncid, sheatid, berg_grid%stored_heat(:,:) )
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_put_var stored_heat failed')
      IF( lwp ) WRITE(numout,*) 'file: ',TRIM(filename),' var: stored_heat written'

      iret = NF90_PUT_VAR( ncid, calvid , p_calving(:,:) )
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_put_var calving failed')
      iret = NF90_PUT_VAR( ncid, calvhid, p_calving_hflx(:,:) )
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_put_var calving_hflx failed')
      IF( lwp ) WRITE(numout,*) 'file: ',TRIM(filename),' var: calving written'

      IF ( ASSOCIATED(first_berg) ) THEN

         ! Write variables
         ! sga - just write out the current point of the trajectory

         this => first_berg
         k = 0
         DO WHILE (ASSOCIATED(this))
            pt => this%current_point
            k=k+1

            iret = NF90_PUT_VAR(ncid, numberid, this%number, (/1,k/), (/nkounts,1/) )
            iret = NF90_PUT_VAR(ncid, scaling_id, this%mass_scaling, (/ k /) )

            iret = NF90_PUT_VAR(ncid, lonid, pt%lon, (/ k /) )
            iret = NF90_PUT_VAR(ncid, latid, pt%lat, (/ k /) )
            iret = NF90_PUT_VAR(ncid, xid, pt%xi, (/ k /) )
            iret = NF90_PUT_VAR(ncid, yid, pt%yj, (/ k /) )
            iret = NF90_PUT_VAR(ncid, uvelid, pt%uvel, (/ k /) )
            iret = NF90_PUT_VAR(ncid, vvelid, pt%vvel, (/ k /) )
            iret = NF90_PUT_VAR(ncid, massid, pt%mass, (/ k /) )
            iret = NF90_PUT_VAR(ncid, thicknessid, pt%thickness, (/ k /) )
            iret = NF90_PUT_VAR(ncid, widthid, pt%width, (/ k /) )
            iret = NF90_PUT_VAR(ncid, lengthid, pt%length, (/ k /) )
            iret = NF90_PUT_VAR(ncid, yearid, pt%year, (/ k /) )
            iret = NF90_PUT_VAR(ncid, dayid, pt%day, (/ k /) )
            iret = NF90_PUT_VAR(ncid, mass_of_bits_id, pt%mass_of_bits, (/ k /) )
            iret = NF90_PUT_VAR(ncid, heat_density_id, pt%heat_density, (/ k /) )

            this=>this%next
         ENDDO

      ENDIF ! associated(first_berg)

      ! Finish up
      iret = NF90_CLOSE(ncid)
      IF (iret .ne. NF90_NOERR) CALL ctl_stop('icebergs, write_restart: nf_close failed')

   END SUBROUTINE icebergs_write_restart

END MODULE icbrst
