MODULE icbthm

   !!======================================================================
   !!                       ***  MODULE  icbthm  ***
   !! Ocean physics:  thermodynamics routines for icebergs
   !!======================================================================
   !! History : 3.3.1 !  2010-01  (Martin&Adcroft) Original code
   !!            -    !  2011-03  (Madec)          Part conversion to NEMO form
   !!            -    !                            Removal of mapping from another grid
   !!            -    !  2011-04  (Alderson)       Split into separate modules
   !!            -    !  2011-05  (Alderson)       Use tmask instead of tmask_i
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   thermodynamics : initialise
   !!                    reference for equations - M = Martin + Adcroft, OM 34, 2010
   !!----------------------------------------------------------------------
   USE par_oce        ! NEMO parameters
   USE dom_oce        ! NEMO domain
   USE in_out_manager ! NEMO IO routines, numout in particular
   USE lib_mpp        ! NEMO MPI routines, ctl_stop in particular
   USE phycst         ! NEMO physical constants
   USE sbc_oce

   USE icb_oce        ! define iceberg arrays
   USE icbutl         ! iceberg utility routines
   USE icbdia         ! iceberg budget routines

   IMPLICIT NONE
   PRIVATE

   PUBLIC   thermodynamics ! routine called in xxx.F90 module

CONTAINS

   SUBROUTINE thermodynamics( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE thermodynamics  ***
      !!
      !! ** Purpose :   compute the iceberg thermodynamics.
      !!
      !! ** Method  : - blah blah
      !!----------------------------------------------------------------------
      INTEGER                         ::   kt          ! timestep number, just passed to print_berg
      !
      REAL(wp)                        ::   M, T, W, L, SST, Vol, Ln, Wn, Tn, nVol, IC, Dn
      REAL(wp)                        ::   Mv, Me, Mb, melt, dvo, dva, dM, Ss, dMe, dMb, dMv
      REAL(wp)                        ::   Mnew, Mnew1, Mnew2, heat
      REAL(wp)                        ::   Mbits, nMbits, dMbitsE, dMbitsM, Lbits, Abits, Mbb
      REAL(wp)                        ::   xi, yj, zff, z1_rday, z1_e1e2, zdt, z1_dt, z1_dt_e1e2
      INTEGER                         ::   ii, ij
      TYPE(iceberg)         , POINTER ::   this, next
      TYPE(point)           , POINTER ::   pt
      !!----------------------------------------------------------------------
      !
      z1_rday = 1._wp / rday
      
      ! we're either going to ignore berg fresh water melt flux and associated heat
      ! or we pass it into the ocean, so at this point we set them both to zero,
      ! accumulate the contributions to them from each iceberg in the while loop following
      ! and then pass them (or not) to the ocean
      !
      berg_grid%floating_melt(:,:) = 0._wp
      berg_grid%calving_hflx(:,:)  = 0._wp
    
      this => first_berg
      DO WHILE( associated(this) )
         !
         pt => this%current_point
         knberg = this%number(1)
         CALL interp_flds( pt%xi, pt%e1, pt%uo, pt%ui, pt%ua, pt%ssh_x, &
            &              pt%yj, pt%e2, pt%vo, pt%vi, pt%va, pt%ssh_y, pt%sst, pt%cn, pt%hi, zff )
         !
         SST = pt%sst
         IC  = MIN( 1._wp, pt%cn + rn_sicn_shift )     ! Shift sea-ice concentration       !!gm ???
         M   = pt%mass
         T   = pt%thickness                               ! total thickness
       ! D   = (rn_rho_bergs/rho_seawater)*T ! draught (keel depth)
       ! F   = T - D ! freeboard
         W   = pt%width
         L   = pt%length
         xi  = pt%xi                                      ! position in (i,j) referential
         yj  = pt%yj
         ii  = INT( xi + 0.5 ) - nimpp + 1                    ! t-cell of the berg
         ij  = INT( yj + 0.5 ) - njmpp + 1
         Vol = T * W * L
         zdt = berg_dt   ;   z1_dt = 1._wp / zdt

         ! Environment
         dvo = SQRT( (pt%uvel-pt%uo)**2 + (pt%vvel-pt%vo)**2 )
         dva = SQRT( (pt%ua  -pt%uo)**2 + (pt%va  -pt%vo)**2 )
         Ss  = 1.5 * SQRT( dva ) + 0.1 * dva                ! Sea state      (eqn M.A9)

         ! Melt rates in m/s (i.e. division by rday)
         Mv = MAX( 7.62e-3*SST+1.29e-3*(SST**2)            , 0._wp ) * z1_rday   ! Buoyant convection at sides (eqn M.A10)
         Mb = MAX( 0.58*(dvo**0.8)*(SST+4.0)/(L**0.2)      , 0._wp ) * z1_rday   ! Basal turbulent melting     (eqn M.A7 )
         Me = MAX( 1./12.*(SST+2.)*Ss*(1+cos(rpi*(IC**3))) , 0._wp ) * z1_rday   ! Wave erosion                (eqn M.A8 )

         IF( ln_operator_splitting ) THEN      ! Operator split update of volume/mass
            Tn    = MAX( T - Mb*zdt , 0._wp )         ! new total thickness (m)
            nVol  = Tn * W * L                        ! new volume (m^3)
            Mnew1 = (nVol/Vol) * M                    ! new mass (kg)
            dMb   = M - Mnew1                         ! mass lost to basal melting (>0) (kg)
            !
            Ln    = MAX( L - Mv*zdt , 0._wp )         ! new length (m)
            Wn    = MAX( W - Mv*zdt , 0._wp )         ! new width (m)
            nVol  = Tn * Wn * Ln                      ! new volume (m^3)
            Mnew2 = (nVol/Vol) * M                    ! new mass (kg)
            dMv   = Mnew1 - Mnew2                     ! mass lost to buoyant convection (>0) (kg)
            !
            Ln    = MAX( Ln - Me*zdt , 0._wp )        ! new length (m)
            Wn    = MAX( Wn - Me*zdt , 0._wp )        ! new width (m)
            nVol  = Tn * Wn * Ln                      ! new volume (m^3)
            Mnew  = ( nVol / Vol ) * M                ! new mass (kg)
            dMe   = Mnew2 - Mnew                      ! mass lost to erosion (>0) (kg)
            dM    = M - Mnew                          ! mass lost to all erosion and melting (>0) (kg)
            !
         ELSE                                         ! Update dimensions of berg
            Ln = MAX( L -(Mv+Me)*zdt ,0._wp )         ! (m)
            Wn = MAX( W -(Mv+Me)*zdt ,0._wp )         ! (m)
            Tn = MAX( T - Mb    *zdt ,0._wp )         ! (m)
            ! Update volume and mass of berg
            nVol = Tn*Wn*Ln                           ! (m^3)
            Mnew = (nVol/Vol)*M                       ! (kg)
            dM   = M - Mnew                           ! (kg)
            dMb = (M/Vol) * (W*   L ) *Mb*zdt         ! approx. mass loss to basal melting (kg)
            dMe = (M/Vol) * (T*(W+L)) *Me*zdt         ! approx. mass lost to erosion (kg)
            dMv = (M/Vol) * (T*(W+L)) *Mv*zdt         ! approx. mass loss to buoyant convection (kg)
         ENDIF

         IF( rn_bits_erosion_fraction > 0._wp ) THEN      ! Bergy bits
            !
            Mbits   = pt%mass_of_bits                                               ! mass of bergy bits (kg)
            dMbitsE = rn_bits_erosion_fraction * dMe                        ! change in mass of bits (kg)
            nMbits  = Mbits + dMbitsE                                               ! add new bergy bits to mass (kg)
            Lbits   = MIN( L, W, T, 40._wp )                                        ! assume bergy bits are smallest dimension or 40 meters
            Abits   = ( Mbits / rn_rho_bergs ) / Lbits                           ! Effective bottom area (assuming T=Lbits)
            Mbb     = MAX( 0.58*(dvo**0.8)*(SST+2.0)/(Lbits**0.2), 0.) * z1_rday    ! Basal turbulent melting (for bits)
            Mbb     = rn_rho_bergs * Abits * Mbb                                 ! in kg/s
            dMbitsM = MIN( Mbb*zdt , nMbits )                                       ! bergy bits mass lost to melting (kg)
            nMbits  = nMbits-dMbitsM                                                ! remove mass lost to bergy bits melt
            IF( Mnew == 0._wp ) THEN                                                ! if parent berg has completely melted then
               dMbitsM = dMbitsM + nMbits                                           ! instantly melt all the bergy bits
               nMbits  = 0._wp
            ENDIF
         ELSE                                                     ! No bergy bits
            Abits   = 0._wp
            dMbitsE = 0._wp
            dMbitsM = 0._wp
            nMbits  = pt%mass_of_bits                             ! retain previous value incase non-zero
         ENDIF

         ! use tmask rather than tmask_i when dealing with icebergs
         IF( tmask(ii,ij,1) /= 0._wp ) THEN    ! Add melting to the grid and field diagnostics
            z1_e1e2    = 1._wp / e1e2t(ii,ij) * this%mass_scaling
            z1_dt_e1e2 = z1_dt * z1_e1e2
            melt    = ( dM - ( dMbitsE - dMbitsM ) ) * z1_dt   ! kg/s
            berg_grid%floating_melt(ii,ij) = berg_grid%floating_melt(ii,ij) + melt    * z1_e1e2    ! kg/m2/s
            heat = melt * pt%heat_density              ! kg/s x J/kg = J/s
            berg_grid%calving_hflx (ii,ij) = berg_grid%calving_hflx (ii,ij) + heat    * z1_e1e2    ! W/m2
            CALL melt_budget(ii, ij, Mnew, heat, this%mass_scaling, dM, dMbitsE, dMbitsM, dMb, dMe, dMv, z1_dt_e1e2 )
         ELSE
            WRITE(numout,*) 'thermodynamics: berg ',this%number(:),' appears to have grounded  at ',narea,ii,ij
            CALL print_berg( this, kt )
            WRITE(numout,*) 'msk=',tmask(ii,ij,1), e1e2t(ii,ij)
            CALL ctl_stop('thermodynamics', 'berg appears to have grounded!')
         ENDIF

         ! Rolling
         Dn = ( rn_rho_bergs / rho_seawater ) * Tn       ! draught (keel depth)
         IF( Dn > 0._wp .AND. MAX(Wn,Ln) < SQRT( 0.92*(Dn**2) + 58.32*Dn ) ) THEN
            T  = Tn
            Tn = Wn
            Wn = T
         endif

         ! Store the new state of iceberg (with L>W)
         pt%mass         = Mnew
         pt%mass_of_bits = nMbits
         pt%thickness    = Tn
         pt%width        = min(Wn,Ln)
         pt%length       = max(Wn,Ln)

         next=>this%next

!!gm  add a test to avoid over melting ?

         IF( Mnew <= 0._wp ) THEN       ! Delete the berg if completely melted
            CALL delete_iceberg_from_list( first_berg, this )
            !
         ELSE                            ! Diagnose mass distribution on grid
            z1_e1e2                 = 1._wp / e1e2t(ii,ij) * this%mass_scaling
            CALL size_budget(ii, ij, Wn, Ln, Abits, this%mass_scaling, Mnew, nMbits, z1_e1e2)
         ENDIF
         !
         this=>next
         !
      END DO
      
      ! now use melt and associated heat flux in ocean (or not)
      !
      IF(.NOT. ln_passive_mode ) THEN
         emp (:,:) = emp (:,:) - berg_grid%floating_melt(:,:)
         emps(:,:) = emps(:,:) - berg_grid%floating_melt(:,:)
!!       qns (:,:) = qns (:,:) + berg_grid%calving_hflx(:,:)  !! heat flux not yet properly coded
      ENDIF
      !
   END  SUBROUTINE thermodynamics

END MODULE icbthm
