MODULE icbtrj

   !!======================================================================
   !!                       ***  MODULE  icbtrj  ***
   !! Ocean physics:  trajectory I/O routines
   !!======================================================================
   !! History : 3.3.1 !  2010-01  (Martin&Adcroft) Original code
   !!            -    !  2011-03  (Madec)          Part conversion to NEMO form
   !!            -    !                            Removal of mapping from another grid
   !!            -    !  2011-05  (Alderson)       New module to handle trajectory output
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   write_trajectory      :
   !!----------------------------------------------------------------------
   USE par_oce        ! NEMO parameters
   USE dom_oce        ! NEMO ocean domain
   USE phycst         ! NEMO physical constants
   USE lib_mpp        ! NEMO MPI library, lk_mpp in particular
   USE in_out_manager ! NEMO IO, numout in particular
   USE netcdf

   USE icb_oce        ! define iceberg arrays
   USE icbutl         ! iceberg utility routines

   IMPLICIT NONE
   PRIVATE

   INTEGER, PRIVATE   ::   kount_traj
   INTEGER, PRIVATE   ::   n_dim, k_dim
   INTEGER, PRIVATE   ::   trajid
   INTEGER, PRIVATE   ::   numberid, stepid, scaling_id
   INTEGER, PRIVATE   ::   lonid, latid, xid, yid, uvelid, vvelid, massid
   INTEGER, PRIVATE   ::   uoid, void, uaid, vaid, uiid, viid
   INTEGER, PRIVATE   ::   sshxid, sshyid, sstid, cntid, thkid
   INTEGER, PRIVATE   ::   thicknessid, widthid, lengthid
   INTEGER, PRIVATE   ::   yearid, dayid
   INTEGER, PRIVATE   ::   mass_of_bits_id, heat_density_id

   PUBLIC   traj_init           ! routine called in xxx.F90 module
   PUBLIC   traj_write          ! routine called in xxx.F90 module
   PUBLIC   traj_sync           ! routine called in xxx.F90 module
   PUBLIC   traj_end            ! routine called in xxx.F90 module

CONTAINS

   !!-------------------------------------------------------------------------

   SUBROUTINE traj_init( nitend )

      ! local variables
      INTEGER, INTENT( in )                 :: nitend
      ! Local variables
      INTEGER                               :: iret, k
      CHARACTER(len=80)                     :: filename
      TYPE(iceberg), POINTER                :: this
      TYPE(point)  , POINTER                :: pt

      IF( lk_mpp ) THEN
         WRITE(filename,'("trajectory_icebergs_",I6.6,"_",I4.4,".nc")') nitend, narea-1
      ELSE
         WRITE(filename,'("trajectory_icebergs_",I6.6,".nc")') nitend
      ENDIF
      IF ( lwp .AND. nn_verbose_level >= 0) WRITE(numout,'(2a)') 'icebergs, traj_init: creating ',TRIM(filename)

      iret = NF90_CREATE(TRIM(filename), NF90_CLOBBER, trajid)
      IF (iret .NE. NF90_NOERR) CALL ctl_stop('icebergs, traj_init: nf_create failed')

      ! Dimensions
      iret = NF90_DEF_DIM(trajid, 'n', NF90_UNLIMITED, n_dim)
      IF (iret .NE. NF90_NOERR) CALL ctl_stop('icebergs, traj_init: nf_def_dim n failed')
      iret = NF90_DEF_DIM(trajid, 'k', nkounts, k_dim)
      IF (iret .NE. NF90_NOERR) CALL ctl_stop('icebergs, traj_init: nf_def_dim k failed')

      ! Variables
      iret = NF90_DEF_VAR(trajid, 'iceberg_number', NF90_INT, (/k_dim,n_dim/), numberid)
      iret = NF90_DEF_VAR(trajid, 'timestep', NF90_INT, n_dim, stepid)
      iret = NF90_DEF_VAR(trajid, 'mass_scaling', NF90_DOUBLE, n_dim, scaling_id)
      iret = NF90_DEF_VAR(trajid, 'lon', NF90_DOUBLE, n_dim, lonid)
      iret = NF90_DEF_VAR(trajid, 'lat', NF90_DOUBLE, n_dim, latid)
      iret = NF90_DEF_VAR(trajid, 'xi', NF90_DOUBLE, n_dim, xid)
      iret = NF90_DEF_VAR(trajid, 'yj', NF90_DOUBLE, n_dim, yid)
      iret = NF90_DEF_VAR(trajid, 'uvel', NF90_DOUBLE, n_dim, uvelid)
      iret = NF90_DEF_VAR(trajid, 'vvel', NF90_DOUBLE, n_dim, vvelid)
      iret = NF90_DEF_VAR(trajid, 'uto', NF90_DOUBLE, n_dim, uoid)
      iret = NF90_DEF_VAR(trajid, 'vto', NF90_DOUBLE, n_dim, void)
      iret = NF90_DEF_VAR(trajid, 'uta', NF90_DOUBLE, n_dim, uaid)
      iret = NF90_DEF_VAR(trajid, 'vta', NF90_DOUBLE, n_dim, vaid)
      iret = NF90_DEF_VAR(trajid, 'uti', NF90_DOUBLE, n_dim, uiid)
      iret = NF90_DEF_VAR(trajid, 'vti', NF90_DOUBLE, n_dim, viid)
      iret = NF90_DEF_VAR(trajid, 'ssh_x', NF90_DOUBLE, n_dim, sshyid)
      iret = NF90_DEF_VAR(trajid, 'ssh_y', NF90_DOUBLE, n_dim, sshxid)
      iret = NF90_DEF_VAR(trajid, 'sst', NF90_DOUBLE, n_dim, sstid)
      iret = NF90_DEF_VAR(trajid, 'icnt', NF90_DOUBLE, n_dim, cntid)
      iret = NF90_DEF_VAR(trajid, 'ithk', NF90_DOUBLE, n_dim, thkid)
      iret = NF90_DEF_VAR(trajid, 'mass', NF90_DOUBLE, n_dim, massid)
      iret = NF90_DEF_VAR(trajid, 'thickness', NF90_DOUBLE, n_dim, thicknessid)
      iret = NF90_DEF_VAR(trajid, 'width', NF90_DOUBLE, n_dim, widthid)
      iret = NF90_DEF_VAR(trajid, 'length', NF90_DOUBLE, n_dim, lengthid)
      iret = NF90_DEF_VAR(trajid, 'year', NF90_INT, n_dim, yearid)
      iret = NF90_DEF_VAR(trajid, 'day', NF90_DOUBLE, n_dim, dayid)
      iret = NF90_DEF_VAR(trajid, 'mass_of_bits', NF90_DOUBLE, n_dim, mass_of_bits_id)
      iret = NF90_DEF_VAR(trajid, 'heat_density', NF90_DOUBLE, n_dim, heat_density_id)

      ! Attributes
      iret = NF90_PUT_ATT(trajid, numberid, 'long_name', 'iceberg number on this processor')
      iret = NF90_PUT_ATT(trajid, numberid, 'units', 'count')
      iret = NF90_PUT_ATT(trajid, stepid, 'long_name', 'timestep number kt')
      iret = NF90_PUT_ATT(trajid, stepid, 'units', 'count')
      iret = NF90_PUT_ATT(trajid, lonid, 'long_name', 'longitude')
      iret = NF90_PUT_ATT(trajid, lonid, 'units', 'degrees_E')
      iret = NF90_PUT_ATT(trajid, latid, 'long_name', 'latitude')
      iret = NF90_PUT_ATT(trajid, latid, 'units', 'degrees_N')
      iret = NF90_PUT_ATT(trajid, xid, 'long_name', 'x grid box position')
      iret = NF90_PUT_ATT(trajid, xid, 'units', 'fractional')
      iret = NF90_PUT_ATT(trajid, yid, 'long_name', 'y grid box position')
      iret = NF90_PUT_ATT(trajid, yid, 'units', 'fractional')
      iret = NF90_PUT_ATT(trajid, uvelid, 'long_name', 'zonal velocity')
      iret = NF90_PUT_ATT(trajid, uvelid, 'units', 'm/s')
      iret = NF90_PUT_ATT(trajid, vvelid, 'long_name', 'meridional velocity')
      iret = NF90_PUT_ATT(trajid, vvelid, 'units', 'm/s')
      iret = NF90_PUT_ATT(trajid, uoid, 'long_name', 'ocean u component')
      iret = NF90_PUT_ATT(trajid, uoid, 'units', 'm/s')
      iret = NF90_PUT_ATT(trajid, void, 'long_name', 'ocean v component')
      iret = NF90_PUT_ATT(trajid, void, 'units', 'm/s')
      iret = NF90_PUT_ATT(trajid, uaid, 'long_name', 'atmosphere u component')
      iret = NF90_PUT_ATT(trajid, uaid, 'units', 'm/s')
      iret = NF90_PUT_ATT(trajid, vaid, 'long_name', 'atmosphere v component')
      iret = NF90_PUT_ATT(trajid, vaid, 'units', 'm/s')
      iret = NF90_PUT_ATT(trajid, uiid, 'long_name', 'sea ice u component')
      iret = NF90_PUT_ATT(trajid, uiid, 'units', 'm/s')
      iret = NF90_PUT_ATT(trajid, viid, 'long_name', 'sea ice v component')
      iret = NF90_PUT_ATT(trajid, viid, 'units', 'm/s')
      iret = NF90_PUT_ATT(trajid, sshxid, 'long_name', 'sea surface height gradient from x points')
      iret = NF90_PUT_ATT(trajid, sshxid, 'units', 'm/m')
      iret = NF90_PUT_ATT(trajid, sshyid, 'long_name', 'sea surface height gradient from y points')
      iret = NF90_PUT_ATT(trajid, sshyid, 'units', 'm/m')
      iret = NF90_PUT_ATT(trajid, sstid, 'long_name', 'sea surface temperature')
      iret = NF90_PUT_ATT(trajid, sstid, 'units', 'degC')
      iret = NF90_PUT_ATT(trajid, cntid, 'long_name', 'sea ice concentration')
      iret = NF90_PUT_ATT(trajid, cntid, 'units', 'degC')
      iret = NF90_PUT_ATT(trajid, thkid, 'long_name', 'sea ice thickness')
      iret = NF90_PUT_ATT(trajid, thkid, 'units', 'm')
      iret = NF90_PUT_ATT(trajid, massid, 'long_name', 'mass')
      iret = NF90_PUT_ATT(trajid, massid, 'units', 'kg')
      iret = NF90_PUT_ATT(trajid, thicknessid, 'long_name', 'thickness')
      iret = NF90_PUT_ATT(trajid, thicknessid, 'units', 'm')
      iret = NF90_PUT_ATT(trajid, widthid, 'long_name', 'width')
      iret = NF90_PUT_ATT(trajid, widthid, 'units', 'm')
      iret = NF90_PUT_ATT(trajid, lengthid, 'long_name', 'length')
      iret = NF90_PUT_ATT(trajid, lengthid, 'units', 'm')
      iret = NF90_PUT_ATT(trajid, yearid, 'long_name', 'calendar year')
      iret = NF90_PUT_ATT(trajid, yearid, 'units', 'years')
      iret = NF90_PUT_ATT(trajid, dayid, 'long_name', 'day of year')
      iret = NF90_PUT_ATT(trajid, dayid, 'units', 'days')
      iret = NF90_PUT_ATT(trajid, scaling_id, 'long_name', 'scaling factor for mass of berg')
      iret = NF90_PUT_ATT(trajid, scaling_id, 'units', 'none')
      iret = NF90_PUT_ATT(trajid, mass_of_bits_id, 'long_name', 'mass of bergy bits')
      iret = NF90_PUT_ATT(trajid, mass_of_bits_id, 'units', 'kg')
      iret = NF90_PUT_ATT(trajid, heat_density_id, 'long_name', 'heat density')
      iret = NF90_PUT_ATT(trajid, heat_density_id, 'units', 'J/kg')

      ! End define mode
      iret = NF90_ENDDEF(trajid)

   END SUBROUTINE traj_init

   !!-------------------------------------------------------------------------

   SUBROUTINE traj_write( kt )
      !! for the moment write out each snapshot of positions
      !! later can rewrite so that it is buffered and written out more efficiently

      INTEGER, INTENT( in )                 :: kt
      ! Local variables
      INTEGER                               :: iret, k
      CHARACTER(len=80)                     :: filename
      TYPE(iceberg), POINTER                :: this
      TYPE(point)  , POINTER                :: pt

      ! Write variables
      ! sga - just write out the current point of the trajectory

      this => first_berg
      k = kount_traj
      DO WHILE (ASSOCIATED(this))
         pt => this%current_point
         k=k+1

         iret = NF90_PUT_VAR(trajid, numberid, this%number, (/1,k/), (/nkounts,1/) )
         iret = NF90_PUT_VAR(trajid, stepid, kt, (/ k /) )
         iret = NF90_PUT_VAR(trajid, scaling_id, this%mass_scaling, (/ k /) )

         iret = NF90_PUT_VAR(trajid, lonid, pt%lon, (/ k /) )
         iret = NF90_PUT_VAR(trajid, latid, pt%lat, (/ k /) )
         iret = NF90_PUT_VAR(trajid, xid, pt%xi, (/ k /) )
         iret = NF90_PUT_VAR(trajid, yid, pt%yj, (/ k /) )
         iret = NF90_PUT_VAR(trajid, uvelid, pt%uvel, (/ k /) )
         iret = NF90_PUT_VAR(trajid, vvelid, pt%vvel, (/ k /) )
         iret = NF90_PUT_VAR(trajid, uoid, pt%uo, (/ k /) )
         iret = NF90_PUT_VAR(trajid, void, pt%vo, (/ k /) )
         iret = NF90_PUT_VAR(trajid, uaid, pt%ua, (/ k /) )
         iret = NF90_PUT_VAR(trajid, vaid, pt%va, (/ k /) )
         iret = NF90_PUT_VAR(trajid, uiid, pt%ui, (/ k /) )
         iret = NF90_PUT_VAR(trajid, viid, pt%vi, (/ k /) )
         iret = NF90_PUT_VAR(trajid, sshxid, pt%ssh_x, (/ k /) )
         iret = NF90_PUT_VAR(trajid, sshyid, pt%ssh_y, (/ k /) )
         iret = NF90_PUT_VAR(trajid, sstid, pt%sst, (/ k /) )
         iret = NF90_PUT_VAR(trajid, cntid, pt%cn, (/ k /) )
         iret = NF90_PUT_VAR(trajid, thkid, pt%hi, (/ k /) )
         iret = NF90_PUT_VAR(trajid, massid, pt%mass, (/ k /) )
         iret = NF90_PUT_VAR(trajid, thicknessid, pt%thickness, (/ k /) )
         iret = NF90_PUT_VAR(trajid, widthid, pt%width, (/ k /) )
         iret = NF90_PUT_VAR(trajid, lengthid, pt%length, (/ k /) )
         iret = NF90_PUT_VAR(trajid, yearid, pt%year, (/ k /) )
         iret = NF90_PUT_VAR(trajid, dayid, pt%day, (/ k /) )
         iret = NF90_PUT_VAR(trajid, mass_of_bits_id, pt%mass_of_bits, (/ k /) )
         iret = NF90_PUT_VAR(trajid, heat_density_id, pt%heat_density, (/ k /) )

         this=>this%next
      ENDDO
      IF( lwp .and. nn_verbose_level > 0 ) WRITE(numout,*) 'trajectory write to frame ', k
      kount_traj = k

   END SUBROUTINE traj_write

   !!-------------------------------------------------------------------------

   SUBROUTINE traj_sync()
      ! Local variables
      INTEGER                               :: iret

      ! flush to file
      iret = NF90_SYNC(trajid)
      IF (iret .NE. NF90_NOERR) CALL ctl_stop('icebergs, traj_sync: nf_sync failed')

   END SUBROUTINE traj_sync

   !!-------------------------------------------------------------------------

   SUBROUTINE traj_end()
      ! Local variables
      INTEGER                               :: iret

      ! Finish up
      iret = NF90_CLOSE(trajid)
      IF (iret .NE. NF90_NOERR) CALL ctl_stop('icebergs, traj_end: nf_close failed')

   END SUBROUTINE traj_end

   !!-------------------------------------------------------------------------

END MODULE icbtrj
